if (typeof window === 'undefined') {
  mono = require("./mono.js");
  self = require("sdk/self");
  window = require("sdk/window/utils").getMostRecentBrowserWindow();
  l10nGet = require("sdk/l10n").get;
  tabs = require("sdk/tabs");
  sdk_timers = require("sdk/timers");
  setInterval = sdk_timers.setInterval;
  clearInterval = sdk_timers.clearInterval;
  window.isModule = true;
}

var engine = (function() {
  var var_cache = {
    // helper name
    helperName: undefined,
    // extension version
    version: undefined,
    // engine ready state
    isReady: false,
    // cache userjsdetected state
    userjsDetected: undefined,
    // trackTime for userTrack
    trackTime: undefined,
    // uuid cache
    uuid: undefined,
    isEnable: 1,
    opButton: null
  };

  var defaultPreferences = {
    version: '0',
    enable: 1,
    button: 1,
    lmFileHosting: 1,
    lmMediaHosting: 1,
    moduleYoutube: 1,
    moduleDailymotion: 1,
    moduleVimeo: 1,
    moduleFacebook: 1,
    moduleSoundcloud: 1,
    moduleVkontakte: 1,
    moduleOdnoklassniki: 1,
    moduleMailru: 1,
    moduleShowDownloadInfo: 1,
    ytHideFLV: 0,
    ytHideMP4: 0,
    ytHideWebM: 1,
    ytHide3GP: 1,
    ytHide3D: 1,
    ytHideMP4NoAudio: 1,
    ytHideAudio_MP4: 1,
    vkShowBitrate: 0,
    sovetnikEnabled: 1
  };

  var preferences = {};

  var preferenceMap = {
    youtube: 'moduleYoutube',
    dailymotion: 'moduleDailymotion',
    vimeo: 'moduleVimeo',
    facebook: 'moduleFacebook',
    soundcloud: 'moduleSoundcloud',
    vk: 'moduleVkontakte',
    odnoklassniki: 'moduleOdnoklassniki',
    mailru: 'moduleMailru'
  };

  var onOptionChange = {
    enable: function() {
      setButtonParams();
      if (mono.isGM) {
        _menu.setEnabled(preferences.enable);
        registerMenuCommands(_moduleName, preferences.enable);
      }
      tabListener.start();
    },
    button: function(enabled) {
      operaShowButton(enabled);
    },
    sovetnikEnabled: function() {
      tabListener.start();
    },
    lmFileHosting: function() {
      tabListener.start();
    },
    lmMediaHosting: function() {
      tabListener.start();
    }
  };

  var msgAction = {
    getLanguage: function(message, cb) {
      cb(language);
    },
    getPreference: function(message, cb) {
      var pref = preferences;
      if (mono.isSafari || mono.isGM) {
        pref = mono.extend({}, preferences);
      }
      cb(pref);

      userTrack();
    },
    getVersion: function(message, cb) {
      cb(var_cache.version);
    },
    updateOption: function(message, cb) {
      preferences[message.key] = message.value;
      var obj = {};
      obj[message.key] = message.value;
      mono.storage.set(obj);
      if (onOptionChange[message.key] !== undefined) {
        onOptionChange[message.key](message.value);
      }
    },
    enableExtension: function() {
      preferences.enable = (preferences.enable)?0:1;
      mono.storage.set({enable: preferences.enable});
      onOptionChange.enable();
    },
    downloadFromCurrentPage: function() {},
    reportBug: function() {},
    viaMenu_updateLinks: function() {
      mono.sendMessage({action: 'updateLinks'}, undefined, 'activeTab');
    },
    viaMenu_downloadMP3Files: function() {
      mono.sendMessage({action: 'downloadMP3Files'}, undefined, 'activeTab');
    },
    viaMenu_downloadPlaylist: function() {
      mono.sendMessage({action: 'downloadPlaylist'}, undefined, 'activeTab');
    },
    viaMenu_downloadPhotos: function() {
      mono.sendMessage({action: 'downloadPhotos'}, undefined, 'activeTab');
    },
    showOptions: function() {
      if (mono.isGM) {
        return _options.show();
      }
      var url = 'options.html';
      if (mono.isFF) {
        url = self.data.url(url);
      }
      if (mono.isSafari) {
        url = safari.extension.baseURI + url;
      }
      mono.openTab(url, true);
    },
    getActiveTabUrl: function(message, cb) {
      if (mono.isChrome) {
        chrome.tabs.getSelected(null, function (tab) {
          return cb(tab.url);
        });
      }
      if (mono.isFF) {
        return cb(tabs.activeTab.url);
      }
      if (mono.isSafari) {
        return cb(safari.application.activeBrowserWindow.activeTab.url);
      }
      if (mono.isOpera) {
        var tab = opera.extension.tabs.getFocused();
        return cb(tab.url);
      }
      if (mono.isGM) {
        return cb(location.href);
      }
    },
    popupResize: function(message) {
      if (mono.isOpera && var_cache.opButton !== null) {
        var_cache.opButton.popup.height = message.height;
        var_cache.opButton.popup.width = message.width;
      }
      if (mono.isSafari) {
        safari.extension.popovers[0].height = message.height;
      }
    },
    getActiveTabModuleName: function(message, cb) {
      mono.sendMessage.call({timeout: 2000}, {action: 'getPageId'}, function(moduleName) {
        cb(moduleName);
      }, 'activeTab');
    },
    userjsDetected: function(message, cb) {
      if (var_cache.userjsDetected) {
        return;
      }
      mono.storage.get('userjsDetected', function(storage) {});
    },
    hasSovetnik: function(message, cb) {
      cb(var_cache.hasSovetnik);
    },
    getBrowser: function(message, cb) {
      cb(var_cache.helperName);
    },
    hideDownloadWarning: function(message, cb) {
      if (message.set !== undefined) {
        return mono.storage.set({hideDownloadWarning: message.set});
      }
      mono.storage.get('hideDownloadWarning', function(storage) {
        cb(storage.hideDownloadWarning);
      });
    },
    storage: function(message, cb) {
      if (message.subaction === 'clear') {
        return;
      }
      mono.storage[message.subaction](message.data, cb);
    },
    trackEvent: function(message) {
      trackEvent(message.category, message.event, message.label);
    },
    trackSocial: function(message) {
      trackSocial(message.target, message.event, message.network);
    }
  };

  var msgWaitList = [];

  var onMessage = function(message, cb) {
    if (var_cache.isReady === false) {
      msgWaitList.push(arguments);
      return;
    }
    var action = message.action || message;
    if (msgAction[action] !== undefined) {
      return msgAction[action](message, cb);
    }
    for (var moduleName in engine.modules) {
      var module = engine.modules[moduleName];
      if (module[action] !== undefined) {
        return module[action](message, cb);
      }
    }
    if (utils[action] !== undefined) {
      return utils[action](message, cb);
    }
  };

  var checkSovetnik = function(cb) {
    if (!mono.isFF && !mono.isChrome) {
      return cb();
    }
    var url = 'includes/sovetnik-sf.js';
    if (mono.isFF) {
      try {
        var data = self.data.load(url);
      } catch (e) {
        return cb();
      }
      return cb(data.length > 10);
    }
    mono.ajax({
      url: url,
      mimeType: 'plain/text',
      success: function(data, xhr) {
        cb(data.length > 10);
      },
      error: function() {
        cb();
      }
    });
  };

  var loadSettings = function(cb) {
    var keys = [];
    for (var key in defaultPreferences) {
      keys.push(key);
    }
    mono.storage.get(keys, function(prefs) {
      for (var key in defaultPreferences) {
        var defaultValue = defaultPreferences[key];
        if (prefs[key] === undefined) {
          prefs[key] = defaultValue;
        } else
        if (typeof prefs[key] === 'string' && typeof defaultValue === 'number') {
          var tValue = parseInt(prefs[key]);
          if (!isNaN(tValue)) {
            prefs[key] = tValue;
          }
        }
      }
      preferences = prefs;

      if (mono.isChrome) {
        if (mono.isChromeVersion < 31) {
          preferences.downloads = false;
          preferences.moduleShowDownloadInfo = false;
          preferences.iframeDownload = false;
        } else {
          preferences.downloads = chrome.downloads !== undefined;
          if (preferences.downloads) {
            preferences.moduleShowDownloadInfo = false;
            preferences.iframeDownload = false;
          } else {
            preferences.iframeDownload = true;
          }
        }
      } else {
        preferences.moduleShowDownloadInfo = false;
      }
      if (mono.isGM) {
        // TODO: Uncomment!
        /*
        if (window.GM_download === undefined || false) {
          preferences.iframeDownload = true;
        } else {
          preferences.downloads = true;
        }
        */
        if (getBrowserName() !== 'firefox') {
          preferences.iframeDownload = true;
        }
        preferences.downloads = false;
      }
      if (mono.isFF && var_cache.ffButton) {
        preferences.downloads = true;
      }
      if (cb !== undefined) {
        cb();
      }
    });
  };

  var language = {};

  var loadLanguage = function(cb, forceLocale) {
    var locale, lang;
    var url = '_locales/{locale}/messages.json';
    if (mono.isChrome) {
      locale = chrome.i18n.getMessage('lang');
      url = url.replace('{locale}', forceLocale || locale);
    } else
    if (mono.isFF) {
      locale = l10nGet('lang');
      if (locale === 'lang') {
        locale = window.navigator.language.substr(0,2);
      }
      url = url.replace('{locale}', forceLocale || locale);
      try {
        lang = self.data.load(url);
        lang = JSON.parse(lang);
      } catch (e) {
        if (forceLocale !== undefined) {
          return cb();
        }
        return loadLanguage(cb, 'en');
      }
      for (var key in lang) {
        language[key] = lang[key].message;
      }
      return cb();
    } else
    if (mono.isSafari || mono.isOpera) {
      url = url.replace('{locale}', forceLocale || navigator.language.substr(0,2));
    } else
    if (mono.isGM) {
      var currentLanguage = navigator.language.substr(0,2);
      lang = _languageList[currentLanguage] || _languageList['en'];
      for (var key in lang) {
        language[key] = lang[key].message;
      }
      return cb();
    }
    mono.ajax({
      mimeType: "application/json",
      dataType: 'json',
      url: url,
      success: function(lang) {
        for (var key in lang) {
          language[key] = lang[key].message;
        }
        cb();
      },
      error: function() {
        if (forceLocale !== undefined) {
          return cb();
        }
        loadLanguage(cb, 'en');
      }
    });
  };

  var getBrowserName = function() {
    var browser = '';
    if(navigator && navigator.userAgent) {
      //console.log(navigator.userAgent);
      if(navigator.userAgent.indexOf('YaBrowser\/') !== -1) {
        browser = 'yabrowser';
      } else
      if(navigator.userAgent.indexOf('OPR\/') !== -1) {
        browser = 'opera-chromium';
      } else
      if(navigator.userAgent.indexOf('Opera\/') !== -1) {
        browser = 'opera';
      } else
      if(navigator.userAgent.indexOf('Firefox\/') !== -1) {
        browser = 'firefox';
      } else
      if(navigator.userAgent.indexOf('Chrome\/') !== -1) {
        browser = 'chrome';
      } else
      if(navigator.userAgent.indexOf('Safari\/') !== -1) {
        browser = 'safari';
      }
    }
    return browser;
  };

  var getHelperName = function() {
    if (mono.isChrome) {
      var browser = getBrowserName();
      if (browser) {
        return browser;
      }

      return 'chrome';
    }
    if (mono.isFF) {
      if (var_cache.hasSovetnik) {
        return 'firefox-sf';
      }
      return 'firefox';
    }
    if (mono.isSafari) {
      return 'safari';
    }
    if (mono.isOpera) {
      return 'opera';
    }
    if (mono.isGM) {
      return 'userjs-' + getBrowserName();
    }

    return 'undefined';
  };

  var setButtonParams = function() {
    if (preferences.enable === var_cache.isEnable) {
      return;
    }
    var_cache.isEnable = preferences.enable;
    if (mono.isGM) {
      return;
    }
    if (mono.isChrome) {
      var icon = {path: 'img/icon_18.png'};
      var title = {title: language.titleDefault};

      if (!preferences.enable) {
        title.title = language.titleDesabled;
        icon.path = 'img/icon_disabled_18.png';
      }

      chrome.browserAction.setIcon(icon);
      chrome.browserAction.setTitle(title);
    }
    if (mono.isFF) {
      if (!var_cache.ffButton) {
        return;
      }
      if (preferences.enable) {
        var_cache.ffButton.icon = {
          '16': self.data.url('img/icon_16.png'),
          '32': self.data.url('img/icon_32.png'),
          '64': self.data.url('img/icon_64.png')
        };
        var_cache.ffButton.label = language.titleDefault;
      } else {
        var_cache.ffButton.icon = {
          '16': self.data.url('img/icon_disabled_18.png'),
          '32': self.data.url('img/icon_disabled_18.png'),
          '64': self.data.url('img/icon_disabled_18.png')
        };
        var_cache.ffButton.label = language.titleDesabled;
      }
      if (var_cache.ffButton.update) {
        var_cache.ffButton.update(var_cache.ffButton);
      }
    }
    if (mono.isSafari) {
      if (preferences.enable) {
        safari.extension.toolbarItems[0].image = safari.extension.baseURI + 'Icon-16.png';
        safari.extension.toolbarItems[0].toolTip = language.titleDefault;
      } else {
        safari.extension.toolbarItems[0].image = safari.extension.baseURI + 'Icon-16-disabled.png';
        safari.extension.toolbarItems[0].toolTip = language.titleDesabled;
      }
    }
    if (mono.isOpera && var_cache.opButton !== null) {
      if (preferences.enable) {
        var_cache.opButton.title = language.titleDefault;
        var_cache.opButton.icon = 'img/icon_18.png';
      } else {
        var_cache.opButton.title = language.titleDesabled;
        var_cache.opButton.icon = 'img/icon_disabled_18.png';
      }
    }
  };

  var getHelperVersion = function(cb) {
    if (mono.isChrome) {
      cb(chrome.app.getDetails().version);
    } else
    if (mono.isFF) {
      cb(self.version);
    }
    if (mono.isSafari) {
      mono.ajax({
        url: safari.extension.baseURI + 'Info.plist',
        success: function(data, xhr) {
          var parser=new DOMParser();
          var xmlDoc=parser.parseFromString(xhr.responseText,"text/xml");
          var elList = xmlDoc.getElementsByTagName('key');
          for (var i = 0, el; el = elList[i]; i++) {
            if (el.textContent === 'CFBundleShortVersionString') {
              return cb(el.nextElementSibling.textContent);
            }
          }
        },
        error: function() {
          cb('unknown');
        }
      });
    }
    if (mono.isOpera) {
      cb(widget.version);
    }
    if (mono.isGM) {
      var version = 'undefined';
      if(GM_info.script) {
        version = GM_info.script.version;
      }
      cb(version);
    }
  };

  var operaShowButton = function(enabled) {
    if (var_cache.opButton !== null) {
      opera.contexts.toolbar.removeItem(var_cache.opButton);
      var_cache.opButton = null;
    }
    if (!enabled) {
      return;
    }
    var_cache.opButton = opera.contexts.toolbar.createItem({});
    setButtonParams();
    opera.contexts.toolbar.addItem(var_cache.opButton);
  };

  var init = function(addon, button, monoLib) {
    if (addon !== undefined) {
      var_cache.monoLib = monoLib;
      mono = mono.init(addon);
      engine.modules.vimeo = require('./vimeo_com_embed.js').init(mono);
      engine.modules.dailymotion = require('./dailymotion_com_embed.js').init(mono);
      engine.modules.youtube = require('./youtube_com_embed.js').init(mono, engine, {platform: window.navigator.platform});
      engine.modules.soundcloud = require('./soundcloud_com_embed.js').init(mono);
      engine.modules.vkontakte = require('./vkontakte_ru_embed.js').init(mono, engine);
      engine.modules.odnoklassniki = require('./odnoklassniki_ru_embed.js').init(mono);
      utils = require('./utils.js').init(mono);
      var_cache.ffButton = button;
    }

    mono.pageId = 'bg';

    mono.onMessage.call({filter: 'bg'}, function (message, response) {
      if (Array.isArray(message)) {
        var countWait = message.length;
        var countReady = 0;
        var resultList = {};
        var ready = function (key, data) {
          countReady += 1;
          resultList[key] = data;
          if (countWait === countReady) {
            response(resultList);
          }
        };
        message.forEach(function (msg) {
          onMessage(msg, function (data) {
            ready(msg, data);
          });
        });
        return;
      }
      onMessage(message, response);
    });
    mono.externalStorageActivate();

    checkSovetnik(function(state) {
      var_cache.hasSovetnik = !!state;
      var_cache.helperName = getHelperName();
      loadLanguage(function() {
        localStorageMigrate(function() {
          loadSettings(function () {
            setButtonParams();
            getHelperVersion(function(version) {
              var_cache.version = version;

              preferences.helper = var_cache.helperName;
              if (preferences.helper === 'firefox-sf') {
                preferences.helper = 'ff-sf';
              } else
              if (preferences.helper === 'firefox') {
                preferences.helper = 'ff';
              }

              if (preferences.enable) {
                checkVersion();
              }
              var_cache.isReady = true;
              msgWaitList.forEach(function(args) {
                onMessage.apply(null, args);
              });
              msgWaitList = [];
            });

            if (mono.isGM && !mono.isIframe()) {
              registerMenuCommands(_moduleName, preferences.enable);

              if(preferences[preferenceMap[_moduleName]] || _moduleName === 'savefrom') {
                _menu.setTitle(language.extName, var_cache.version);
                _menu.create(preferences.enable);
              }
            }

            if (mono.isOpera) {
              operaShowButton(preferences.button);
            }

            tabListener.start();
          });
        });
      });
    });

    if (mono.isSafari) {
      safari.extension.settings.addEventListener('change', function(event){
        if (event.key !== 'show_options') {
          return;
        }
        var window = safari.application.activeBrowserWindow;
        var tab = window.openTab();
        tab.url = safari.extension.baseURI + 'options.html';
        tab.activate();
      });
    }
  };

  var tabListener = {
    listen: false,
    state: false,
    chMessageSender: undefined,
    excludeList: [
      "*.google.*",
      "*.acidtests.org",

      "*.savefrom.net",
      "*.youtube.com",
      "*.vimeo.com",
      "*.dailymotion.*",
      "*.vk.com",
      "*.vkontakte.ru",
      "*.odnoklassniki.ru",
      "my.mail.ru",
      "*.ok.ru",
      "*.soundcloud.com",
      "*.facebook.com"
    ],
    matchCache: undefined,
    createMatchCache: function() {
      var rList = [];
      for (var i = 0, item; item = tabListener.excludeList[i]; i++) {
        var regexp = item.replace(/\./g, '\\.').replace(/\*\\\./, '\\.?').replace(/\.\*/g, '..?');
        if (item.substr(item.length-1, 1) !== '*') {
          regexp += '$';
        }
        if (item[0] !== '*') {
          regexp = '^'+regexp;
        }
        rList.push(regexp);
      }
      tabListener.matchCache = new RegExp(rList.join('|'), 'i');
    },
    checkUrl: function(url) {
      if (url.indexOf('http') !== 0) {
        return;
      }
      var dStart = url.indexOf('://') + 3;
      if (dStart === -1) {
        return;
      }
      var dEnd = url.indexOf('/', dStart);
      if (dEnd === -1) {
        dEnd = url.length - 1;
      }
      var dName = url.substr( dStart, dEnd - dStart );

      if (tabListener.matchCache.test(dName)) {
        return;
      }
      return 1;
    },
    onChange: function(tab, tabId) {
      if (mono.isChrome) {
        chrome.tabs.executeScript(tabId, { file: 'js/mono.js' }, function() {
          if (preferences.lmFileHosting || preferences.lmMediaHosting) {
            chrome.tabs.executeScript(tabId, { file: 'includes/components.js' }, function() {
              chrome.tabs.executeScript(tabId, { file: 'includes/link_modifier.js' });
            });
          }
          if (var_cache.hasSovetnik && preferences.sovetnikEnabled) {
            chrome.tabs.executeScript(tabId, { file: 'includes/sovetnik-sf.js' });
          }
        });
      } else
      if (mono.isFF) {
        var list = [self.data.url('js/mono.js')];

        if (preferences.lmFileHosting || preferences.lmMediaHosting) {
          list.push(self.data.url('includes/components.js'));
          list.push(self.data.url('includes/link_modifier.js'));
        }
        if (var_cache.hasSovetnik && preferences.sovetnikEnabled) {
          list.push(self.data.url('includes/sovetnik-sf.js'));
        }
        
        var worker = tab.attach({
          contentScriptFile: list
        });
        var_cache.monoLib.addPage('lm', worker);
      }
    },
    start: function() {
      if (!mono.isChrome && !mono.isFF) {
        return;
      }

      if (!preferences.enable ||
        (!preferences.sovetnikEnabled && !preferences.lmFileHosting && !preferences.lmMediaHosting)) {
        tabListener.state = false;
        return;
      }

      if (tabListener.state) {
        return;
      }
      tabListener.state = true;

      if (tabListener.listen) {
        return;
      }
      if (mono.isChrome) {
        if (chrome.runtime && chrome.runtime.onMessage) {
          tabListener.chMessageSender = chrome.tabs.sendMessage;
        } else {
          tabListener.chMessageSender = chrome.tabs.sendRequest;
        }
      }

      tabListener.createMatchCache();
      if (mono.isChrome) {
        chrome.tabs.onUpdated.addListener(function chListener(tabId, changeInfo, tab) {
          if (tabListener.state === false) {
            chrome.tabs.onUpdated.removeListener(chListener);
            tabListener.listen = false;
            return;
          }
          if (changeInfo.status !== 'complete') { // complete or loading
            return;
          }
          if (tabListener.checkUrl(tab.url) === undefined) {
            return;
          }

          tabListener.chMessageSender(tabId, 'hasInject', function(r) {
            if (r !== undefined) {
              return;
            }
            tabListener.onChange(tab, tabId);
          });
        });
      } else
      if (mono.isFF) {
        tabs.on('ready', function ffListener(tab) {
          if (tabListener.state === false) {
            tabs.removeListener('ready', ffListener);
            tabListener.listen = false;
            return;
          }
          if (tabListener.checkUrl(tab.url) === undefined) {
            return;
          }
          tabListener.onChange(tab);
        });
      }
      tabListener.listen = true;
    }
  };

  function dblTrackCheck(cb) {
    if (!mono.isGM) {
      return cb();
    }
    mono.storage.get('dblTrack', function(storage) {
      var dblTrack = storage.dblTrack;
      var dataList = (dblTrack)?dblTrack.split(','):'';
      var now = Date.now();
      if (dataList && dataList[1] > now) {
        return;
      }
      var uuid = generateUuid();
      var expire = now + 60000;
      mono.storage.set({dblTrack: uuid+','+expire});

      setTimeout(function() {
        mono.storage.get('dblTrack', function(storage) {
          var dblTrack = storage.dblTrack;
          var dataList = (dblTrack)?dblTrack.split(','):'';
          if (dataList[0] !== uuid) {
            return;
          }
          cb();
        });
      }, 5000);
    });
  }

  function userTrack() {
    var checkTrackTime = function(currentTime) {
      var now = Date.now();
      if (currentTime > now) {
        return;
      }
      var_cache.trackTime = now + 86400000;
      mono.storage.set({ trackTime: String(var_cache.trackTime) });
      dblTrackCheck(function() {
        sendScreenViewStats('init');
      });
    };

    if (var_cache.trackTime !== undefined) {
      return checkTrackTime(var_cache.trackTime);
    }

    mono.storage.get('trackTime', function(storage) {
      var_cache.trackTime = storage.trackTime;
      return checkTrackTime(var_cache.trackTime);
    });
  }

  function sendScreenViewStats(screenName) {
    var params = {
      an: 'helper',
      aid: var_cache.helperName,
      av: var_cache.version,
      t: 'screenview',
      cd: screenName
    };

    sendStatsInfo(params);
  }

  function trackSocial(target, action, network) {
    var params = {
      st: target, // /home
      sa: action, // like
      sn: network, // facebook
      t: 'social'
    };

    sendStatsInfo(params);
  }

  function trackEvent(category, action, label) {
    var params = {
      ec: category, // share-button
      ea: action, // click
      el: label, // vk
      t: 'event'
    };

    sendStatsInfo(params);
  }

  function sendStatsInfo(params) {
    if(!params.t) {
      return;
    }
    getUuid(function(uuid) {
      var def = {
        v: 1,
        tid: 'UA-7055055-5',
        cid: uuid
        //ul: chrome.i18n.getMessage("lang")
      };

      for (var i in def) {
        if(!params[i])
          params[i] = def[i];
      }

      mono.ajax({
        url: 'https://www.google-analytics.com/collect',
        type: 'POST',
        data: mono.param(params)
      });
    });
  }

  function getUuid(cb) {
    if (var_cache.uuid !== undefined) {
      return cb(var_cache.uuid);
    }

    mono.storage.get('uuid', function(storage) {
      if (!storage.uuid) {
        storage.uuid = generateUuid();
        mono.storage.set(storage);
      }
      var_cache.uuid = storage.uuid;
      cb(var_cache.uuid);
    });
  }

  function generateUuid()
  {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c) {
      var r = Math.random()*16|0, v = c == 'x' ? r : (r&0x3|0x8);
      return v.toString(16);
    });
  }

  var checkVersion = function () {
    var url, needSave;

    if(preferences.version === '0') {} else
    if(preferences.version !== var_cache.version) {}

    if (needSave) {
      msgAction.updateOption({key: 'version', value: var_cache.version});
    }

    if (mono.isGM) {
      return;
    }

    if(url !== undefined) {
      checkUrlsOfOpenTabs([],
        function(foundUrls) {
          if(foundUrls.length > 0) {
            return;
          }

          checkUrlsOfOpenTabs([/https?:\/\/legal\.yandex\.(ru|com\.tr)\//i], function(foundUrls){
            var active = foundUrls.length === 0;
            mono.openTab(url, active, active);
          });
        });
      }
  };

  var checkUrlsOfOpenTabs = function(regExpList, callback) {
    var urlList = [];
    var checkUrlList = function() {
      var foundUrlList = [];
      urlList.forEach(function(url) {
        regExpList.forEach(function(regexp) {
          if (url.search(regexp) !== -1 ) {
            foundUrlList.push(url);
          }
        });
      });
      callback(foundUrlList);
    };
    if (mono.isChrome) {
      chrome.tabs.query({}, function (tabs) {
        tabs.forEach(function (tab) {
          urlList.push(tab.url);
        });
        checkUrlList();
      });
    } else
    if (mono.isFF) {
      for (var tab in tabs) {
        urlList.push(tabs[tab].url);
      }
      checkUrlList();
    } else
    if (mono.isOpera) {
      var oTabs = opera.extension.tabs.getAll();
      oTabs.forEach(function(tab) {
        urlList.push(tab.url);
      });
      checkUrlList();
    } else
    if (mono.isSafari) {
      safari.application.activeBrowserWindow.tabs.forEach(function(tab) {
        if (!tab.url) {
          return 1;
        }
        urlList.push(tab.url);
      });
      checkUrlList();
    } else
    if (mono.isGM) {
      urlList.push( location.href );
      checkUrlList();
    }
  };

  var localStorageMigrate = function(cb) {
    var lStorage;
    if (mono.isOpera) {
      lStorage = widget.preferences;
    } else {
      if (typeof localStorage === 'undefined') {
        return cb();
      }
      lStorage = localStorage;
    }
    if (lStorage.length === 0) {
      return cb();
    }
    mono.storage.get('migrated', function(storage) {
      if (storage.migrated !== undefined) {
        return cb();
      }
      var obj = {migrated: 1};
      var copyDP = mono.extend({}, defaultPreferences);
      copyDP.helper = '';
      copyDP.uuid = '';
      copyDP.trackTime = '';
      copyDP.dblTrack = '';
      for (var key in copyDP) {
        var value = lStorage[key];
        if (!value) {
          continue;
        }
        if (typeof copyDP[key] === 'number') {
          obj[key] = parseInt(value);
          if (isNaN(obj[key])) {
            obj[key] = copyDP[key];
          }
        } else {
          obj[key] = value;
        }
      }
      mono.storage.set(obj, function() {
        cb();
      });
    });
  };

  var registerMenuCommands = (function () {
    var menuCommands = [
      {
        id: 'downloadFromCurrentPage',
        command: function() {
          msgAction.downloadFromCurrentPage();
        },
        modules: ['lm', 'youtube', 'soundcloud', 'dailymotion', 'vimeo']
      }, {
        id: 'updateLinks',
        notify: 'updateLinksNotification',
        modules: ['vk', 'odnoklassniki', 'facebook', 'lm']
      }, {
        id: 'downloadMP3Files',
        modules: ['vk', 'mailru']
      }, {
        id: 'downloadPlaylist',
        modules: ['vk', 'mailru']
      }, {
        id: 'downloadPhotos',
        modules: ['vk']
      }, {
        id: 'showOptions',
        command: function() {
          _options.show();
        },
        always: true
      }, {
        id: 'reportBug',
        command: function() {
          msgAction.reportBug();
        },
        always: true
      }, {
        id: 'disable',
        command: function() {
          msgAction.enableExtension();
        },
        always: true
      }
    ];

    var registerModuleMenuCommand = function (params, enabled) {
      if(!params.id) {
        // separator
        return;
      }

      var strId = params.id;
      if(!enabled && params.id == 'disable') {
        strId = 'menuEnable';
      }

      var name = language[strId];
      if(!name) {
        return;
      }

      if(params.command) {
        GM_registerMenuCommand(name, params.command);
        return;
      }

      var fn = function() {
        onMessage({action: ( params.modules !== undefined ? 'viaMenu_' : '' ) + params.id});

        if(params.notify) {
          GM_notification(language[params.notify], null, null, null, 3000);
        }
      };

      GM_registerMenuCommand(name, fn);
    };

    return function (moduleName, enabled) {
      for (var i = 0; i < menuCommands.length; i++) {
        var item = menuCommands[i];

        if (item.gmId) {
          continue;
          GM_unregisterMenuCommand(item.gmId);
        }

        if (enabled || item.always) {
          if (!item.modules || item.modules.indexOf(moduleName) > -1) {
            menuCommands[i].gmId = registerModuleMenuCommand(item, enabled);
          }
        }
      }
    }
  }());

  return {
    init: init,
    modules: {},
    language: language,
    onMessage: onMessage
  }
})();

if (window.isModule) {
  exports.init = engine.init;
} else
if (!mono.isGM) {
  engine.init();
}