// ==UserScript==
// @name        Mono
//
// @exclude     file://*
// @exclude     http://google.*/*
// @exclude     http://*.google.*/*
// @exclude     https://google.*/*
// @exclude     https://*.google.*/*
//
// ==/UserScript==

/**
 * Created by Anton on 21.06.2014.
 *
 * Mono cross-browser engine.
 */
var mono = function (env) {
  /**
   * @namespace chrome
   * @namespace chrome.app
   * @namespace chrome.app.getDetails
   *
   * @namespace safari.self.identifier
   *
   * @namespace addon
   */
  var mono = function() {
    // mono like console.log
    var args = Array.prototype.slice.call(arguments);
    args.unshift(mono.pageId,'monoLog:');
    console.log.apply(console, args);
  };

  var checkCompatibility = function() {
    if (mono.isTM || mono.isChrome) {
      var vPos = navigator.userAgent.indexOf('Chrome/') + 7;
      var vPosEnd = navigator.userAgent.indexOf('.', vPos);
      var chromeVersion = parseInt(navigator.userAgent.substr(vPos, vPosEnd - vPos));
      if (isNaN(chromeVersion)) {
        return;
      }
      mono.isChromeVersion = chromeVersion;
      if (chromeVersion < 31) {
        mono.noMouseEnter = true;
        mono.noXhrJson = true;
      }
    }
  };

  var defaultId = 'monoScope';
  var addon = undefined;
  if (typeof window === 'undefined') {
    mono.isModule = true;
    mono.isFF = true;
    addon = env;
  } else {
    if (typeof GM_getValue !== 'undefined') {
      mono.isGM = true;
      if (window.chrome !== undefined) {
        mono.isTM = true;
        checkCompatibility();
      }
    } else
    if (window.chrome !== undefined) {
      mono.isChrome = true;
      if (chrome.app.getDetails === undefined) {
        mono.isChromeApp = true;
      } else {
        var details = chrome.app.getDetails();
        if (details && details.app !== undefined) {
          mono.isChromeWebApp = true;
        }
      }
      mono.isChromeInject = chrome.tabs === undefined;
      checkCompatibility();
    } else
    if (window.safari !== undefined) {
      mono.isSafari = true;
      mono.isSafariPopup = safari.self.identifier === 'popup';
      mono.isSafariBgPage = safari.self.addEventListener === undefined;
      mono.isSafariInject = !mono.isSafariPopup && safari.application === undefined;
      mono.noMouseEnter = true;
      mono.noXhrJson = true;
    } else
    if (window.opera !== undefined) {
      mono.isOpera = true;
      mono.isOperaInject = opera.extension.broadcastMessage === undefined;
      mono.noMouseEnter = true;
    } else {
      addon = window.addon || window.self;
      if (addon !== undefined && addon.port !== undefined) {
        mono.isFF = true;
      } else
      if (navigator.userAgent.indexOf('Firefox') !== -1) {
        mono.isFF = true;
        mono.noAddon = true;
      }
    }
  }
  mono.messageStack = 50;
  mono.addon = addon;
  mono.pageId = defaultId;
  mono.debug = {
    messages: false
  };

  var messagesEnable = false;

  var externalStorage = {
    get: function(obj, cb) {
      mono.sendMessage({action: 'get', data: obj}, cb, 'monoStorage');
    },
    set: function(obj, cb) {
      mono.sendMessage({action: 'set', data: obj}, cb, 'monoStorage');
    },
    remove: function(obj, cb) {
      mono.sendMessage({action: 'remove', data: obj}, cb, 'monoStorage');
    },
    clear: function(cb) {
      mono.sendMessage({action: 'clear'}, cb, 'monoStorage');
    }
  };

  mono.externalStorageActivate = function() {
    if (!mono.isChrome && !mono.isSafari && !mono.isOpera) {
      return;
    }
    !messagesEnable && mono.onMessage(function(){});
    mono.onMessage.call({filter: 'monoStorage', private: true}, function(message, response) {
      if (message.action === 'get') {
        return mono.storage.get(message.data, response);
      } else
      if (message.action === 'set') {
        return mono.storage.set(message.data, response);
      } else
      if (message.action === 'remove') {
        return mono.storage.remove(message.data, response);
      } else
      if (message.action === 'clear') {
        return mono.storage.clear(response);
      }
    })
  };

  var localStorage = {};
  var localStorageMode = {
    getObj: function(key) {
      var index = 0;
      var keyPrefix = localStorageMode.chunkPrefix+key;
      var chunk = localStorage[keyPrefix+index];
      var data = '';
      while (chunk !== undefined) {
        data += chunk;
        index++;
        chunk = localStorage[keyPrefix+index];
      }
      var value = undefined;
      try {
        value = JSON.parse(data);
      } catch (e) {}
      return value;
    },
    setObj: function(key, value) {
      value = JSON.stringify(value);
      var keyPrefix = localStorageMode.chunkPrefix+key;
      var chunkLen = 1024 - keyPrefix.length - 3;
      if (localStorageMode.regexp === undefined) {
        localStorageMode.regexp = new RegExp('.{1,'+chunkLen+'}', 'g');
      }
      var valueLen = value.length;
      var number_of_part = Math.floor( valueLen / chunkLen );
      if (number_of_part >= 512) {
        console.log(mono.pageId, 'monoLog:', 'localStorage','Can\'t save item', key,', very big!');
        return;
      }
      var dataList = value.match(localStorageMode.regexp);
      var dataListLen = dataList.length;
      for (var i = 0, item; i < dataListLen; i++) {
        item = dataList[i];
        localStorage[keyPrefix+i] = item;
      }
      localStorage[key] = localStorageMode.chunkItem;

      localStorageMode.rmObj(key, dataListLen);
    },
    rmObj: function(key, index) {
      var keyPrefix = localStorageMode.chunkPrefix+key;
      if (index === undefined) {
        index = 0;
      }
      var data = localStorage[keyPrefix+index];
      while (data !== undefined) {
        delete localStorage[keyPrefix+index];
        index++;
        data = localStorage[keyPrefix+index];
      }
    },
    readValue: function(key, value) {
      if (value === localStorageMode.chunkItem) {
        value = localStorageMode.getObj(key)
      } else
      if (value !== undefined) {
        var type = localStorage['_keyType_'+key];
        if (type === 'boolean') {
          value = value === 'true';
        } else
        if (type !== 'string') {
          value = parseInt(value);
        }
      }
      return value;
    },
    get: function (src, cb) {
      var key, obj = {};
      if (src === undefined || src === null) {
        for (key in localStorage) {
          if (!localStorage.hasOwnProperty(key) || key === 'length') {
            continue;
          }
          if (key.substr(0, localStorageMode.chunkLen) === localStorageMode.chunkPrefix) {
            continue;
          }
          if (key.substr(0, 9) === '_keyType_') {
            continue;
          }
          obj[key] = localStorageMode.readValue(key, localStorage[key]);
        }
        return cb(obj);
      }
      if (typeof src === 'string') {
        src = [src];
      }
      if (Array.isArray(src) === true) {
        for (var i = 0, len = src.length; i < len; i++) {
          key = src[i];
          obj[key] = localStorageMode.readValue(key, localStorage[key]);
        }
      } else {
        for (key in src) {
          obj[key] = localStorageMode.readValue(key, localStorage[key]);
        }
      }
      cb(obj);
    },
    set: function (obj, cb) {
      var key;
      for (key in obj) {
        var value = obj[key];
        if (value === undefined) {
          localStorageMode.remove(key);
        } else
        if (typeof value === 'object') {
          localStorageMode.setObj(key, value);
        } else {
          if (typeof value === 'boolean') {
            localStorage['_keyType_'+key] = 'boolean';
          } else
          if (typeof value !== 'number') {
            localStorage['_keyType_'+key] = 'string';
          } else {
            delete localStorage['_keyType_'+key];
          }
          localStorage[key] = value;
        }
      }
      cb && cb();
    },
    remove: function (obj, cb) {
      if (Array.isArray(obj)) {
        for (var i = 0, len = obj.length; i < len; i++) {
          var key = obj[i];
          if (localStorage[key] === localStorageMode.chunkItem) {
            localStorageMode.rmObj(key);
          }
          delete localStorage[key];
        }
      } else {
        if (localStorage[obj] === localStorageMode.chunkItem) {
          localStorageMode.rmObj(obj);
        }
        delete localStorage[obj];
      }
      cb && cb();
    },
    clear: function (cb) {
      localStorage.clear();
      cb && cb();
    }
  };
  localStorageMode.chunkPrefix = 'mCh_';
  localStorageMode.chunkLen = localStorageMode.chunkPrefix.length;
  localStorageMode.chunkItem = 'monoChunk';

  var monoStorage = function() {
    /**
     * @namespace require
     */
    var ss = require("sdk/simple-storage");
    return {
      get: function (src, cb) {
        var key, obj = {};
        if (src === undefined || src === null) {
          for (key in ss.storage) {
            if (!ss.storage.hasOwnProperty(key)) {
              continue;
            }
            obj[key] = ss.storage[key];
          }
          return cb(obj);
        }
        if (typeof src === 'string') {
          src = [src];
        }
        if (Array.isArray(src) === true) {
          for (var i = 0, len = src.length; i < len; i++) {
            key = src[i];
            obj[key] = ss.storage[key];
          }
        } else {
          for (key in src) {
            obj[key] = ss.storage[key];
          }
        }
        cb(obj);
      },
      set: function (obj, cb) {
        var key;
        for (key in obj) {
          ss.storage[key] = obj[key];
        }
        cb && cb();
      },
      remove: function (obj, cb) {
        if (Array.isArray(obj)) {
          for (var i = 0, len = obj.length; i < len; i++) {
            var key = obj[i];
            delete ss.storage[key];
          }
        } else {
          delete ss.storage[obj];
        }
        cb && cb();
      },
      clear: function (cb) {
        var key;
        for (key in ss.storage) {
          delete ss.storage[key];
        }
        cb && cb();
      }
    }
  };

  var gmStorage = {
    /**
     * @namespace GM_listValues
     * @namespace GM_getValue
     * @namespace GM_setValue
     * @namespace GM_deleteValue
     */
    get: function (src, cb) {
      var key, obj = {};
      if (src === undefined || src === null) {
        var nameList = GM_listValues();
        for (key in nameList) {
          obj[key] = GM_getValue(key);
        }
        return cb(obj);
      }
      if (typeof src === 'string') {
        src = [src];
      }
      if (Array.isArray(src) === true) {
        for (var i = 0, len = src.length; i < len; i++) {
          key = src[i];
          obj[key] = GM_getValue(key);
        }
      } else {
        for (key in src) {
          obj[key] = GM_getValue(key);
        }
      }
      cb(obj);
    },
    set: function (obj, cb) {
      var key;
      for (key in obj) {
        GM_setValue(key, obj[key]);
      }
      cb && cb();
    },
    remove: function (obj, cb) {
      if (Array.isArray(obj)) {
        for (var i = 0, len = obj.length; i < len; i++) {
          var key = obj[i];
          GM_deleteValue(key);
        }
      } else {
        GM_deleteValue(obj);
      }
      cb && cb();
    },
    clear: function (cb) {
      var key;
      var nameList = GM_listValues();
      for (key in nameList) {
        GM_deleteValue(key);
      }
      cb && cb();
    }
  };

  var storage_fn = function(mode) {
    /**
     * @namespace widget.preferences
     */
    if (mono.isModule) {
      if (monoStorage.get === undefined) {
        monoStorage = monoStorage();
      }
      return monoStorage;
    } else
    if (mono.isChrome && chrome.storage !== undefined) {
      return chrome.storage[mode];
    } else
    if (mono.isOpera && window.widget) {
      localStorage = widget.preferences;
      return localStorageMode;
    } else
    if (mono.isGM) {
      return gmStorage;
    } else
    if (mono.isFF || mono.isChromeInject || mono.isOperaInject || mono.isSafariInject) {
      return externalStorage;
    } else
    if (window.localStorage !== undefined) {
      localStorage = window.localStorage;
      return localStorageMode;
    }
    return {};
  };
  mono.storage = storage_fn('local');
  mono.storage.local = mono.storage;
  mono.storage.sync = storage_fn('sync');

  var msgTools = function() {
    var cbObj = mono.debug.cbStack = {};
    var cbStack = [];
    var id = 0;
    var idPrefix = Math.floor(Math.random()*1000)+'_';
    return {
      cbCollector: function (message, cb) {
        if (cb === undefined) {
          return;
        }
        mono.debug.messages && console.log(mono.pageId, 'monoLog:', 'cbCollector', message);
        !messagesEnable && mono.onMessage(function(){});
        if (cbStack.length > mono.messageStack) {
          console.log(mono.pageId, 'monoLog:', 'Stack overflow!');
          delete cbObj[cbStack.shift()];
        }
        id++;
        message.monoCallbackId = idPrefix+id;
        cbObj[idPrefix+id] = cb;
        cbStack.push(idPrefix+id);

        var timeout = message.timeout;
        if (timeout !== undefined) {
          delete message.timeout;
          setTimeout(function() {
            if (cbObj[idPrefix+id] === undefined) {
              return;
            }
            msgTools.cbCaller({
              data: null,
              monoResponseId: idPrefix+id
            });
          }, timeout);
        }
      },
      cbCaller: function(message) {
        mono.debug.messages && console.log(mono.pageId, 'monoLog:', 'cbCaller', message);
        if (cbObj[message.monoResponseId] === undefined) {
          if (message.monoResponseId.indexOf(idPrefix) === -1) {
            return;
          }
          return console.log(mono.pageId, 'monoLog:', 'Send to', message.monoTo, 'Id', message.monoResponseId,'Message response not found!');
        }
        cbObj[message.monoResponseId](message.data);
        delete cbObj[message.monoResponseId];
        cbStack.splice(cbStack.indexOf(message.monoResponseId), 1);
      },
      mkResponse: function(message) {
        if (message.monoCallbackId === undefined) {
          return undefined;
        }
        mono.debug.messages && console.log(mono.pageId, 'monoLog:', 'mkResponse', message);
        return function(responseMessage) {
          if (message.dead === 1) {
            return console.log(mono.pageId, 'monoLog:', 'mkResponse', 'dead response', message);
          }
          responseMessage = {
            data: responseMessage,
            monoResponseId: message.monoCallbackId,
            monoTo: message.monoFrom,
            monoFrom: message.monoTo
          };
          if (message.tabId !== undefined) {
            responseMessage.tabId = message.tabId;
          }
          if (message.source !== undefined) {
            responseMessage.source = message.source;
          }
          mono.debug.messages && console.log(mono.pageId, 'monoLog:', 'sendResponse', responseMessage);
          mono.sendMessage.send(responseMessage);
          message.dead = 1;
        };
      },
      rmCaller: function(monoResponseId) {
        if (monoResponseId === undefined) {
          return;
        }
        delete cbObj[monoResponseId];
        cbStack.splice(cbStack.indexOf(monoResponseId), 1);
      },
      readFilter: function(params, cb) {
        cb.private = params.private;
        if (params.filter === undefined) {
          params.filter = [mono.pageId];
        } else
        if (!Array.isArray(params.filter)) {
          params.filter = [params.filter];
        }
        cb.filter = params.filter;
      },
      filter: function(message, cb) {
        if (message.monoTo === defaultId) {
          if (cb.private !== undefined) {
            mono.debug.messages && console.log(mono.pageId, 'monoLog:', 'filterDrop', message);
            return 1;
          }
        } else
        if (cb.filter.indexOf(message.monoTo) === -1) {
          mono.debug.messages && console.log(mono.pageId, 'monoLog:', 'filterDrop', message);
          return 1;
        }
        return undefined;
      }
    }
  }();

  var ffVirtualPort = function() {
    /**
     * @namespace postMessage
     */
    var onCollector = [];
    mono.addon = addon = {
      port: {
        emit: function(pageId, message) {
          var msg = '>'+JSON.stringify(message);
          window.postMessage(msg, "*");
        },
        on: function(pageId, onMessage) {
          onCollector.push(onMessage);
          if (onCollector.length > 1) {
            return;
          }
          window.addEventListener('monoMessage', function (e) {
            if (e.detail[0] !== '<') {
              return;
            }
            var data = e.detail.substr(1);
            var json = JSON.parse(data);
            for (var i = 0, cb; cb = onCollector[i]; i++) {
              cb(json);
            }
          });
        }
      }
    }
  };

  var ffMessaging = {
    cbList: [],
    currentTab: function (message) {
      mono.sendMessage({action: 'toActiveTab', message: message}, undefined, 'service');
    },
    send: function(message) {
      addon.port.emit('mono', message);
    },
    on: function(cb) {
      msgTools.readFilter(this, cb);
      ffMessaging.cbList.push(cb);
      if (ffMessaging.cbList.length > 1) {
        return;
      }
      messagesEnable = true;
      var onMessage = function(message) {
        if (message.monoResponseId !== undefined) {
          return msgTools.cbCaller(message);
        }
        var response = msgTools.mkResponse(message);
        for (var i = 0, itemCb; itemCb = ffMessaging.cbList[i]; i++) {
          if (msgTools.filter(message, itemCb)) {
            continue;
          }
          itemCb(message.data, response);
        }
      };
      addon.port.on('mono', onMessage);
    }
  };

  var chRequestMessaging = {
    cbList: [],
    currentTab: function (message) {
      chrome.tabs.query({active: true, currentWindow: true}, function(tabs) {
        if (tabs[0] === undefined || tabs[0].id < 0) {
          msgTools.rmCaller(message.monoResponseId);
          return;
        }
        message.monoTo = defaultId;
        chrome.tabs.sendRequest(tabs[0].id, message, function(message) {
          if (message && message.monoResponseId) {
            return msgTools.cbCaller(message);
          }
        });
      });
    },
    send: function(message) {
      var tabId = message.tabId;
      if (tabId !== undefined) {
        delete message.tabId;
        return chrome.tabs.sendRequest(tabId, message);
      }
      var source = message.source;
      if (source !== undefined) {
        delete message.source;
        return source(message);
      }
      chrome.extension.sendRequest(message, function(message) {
        if (message && message.monoResponseId) {
          return msgTools.cbCaller(message);
        }
      });
    },
    on: function(cb) {
      msgTools.readFilter(this, cb);
      chRequestMessaging.cbList.push(cb);
      if (chRequestMessaging.cbList.length > 1) {
        return;
      }
      messagesEnable = true;
      chrome.extension.onRequest.addListener(function(message, sender, _response) {
        if (message.monoResponseId) {
          return msgTools.cbCaller(message);
        }
        if (sender.tab && sender.tab.id > -1) {
          message.tabId = sender.tab.id;
        } else {
          message.source = _response;
        }
        var response = msgTools.mkResponse(message);
        for (var i = 0, itemCb; itemCb = chRequestMessaging.cbList[i]; i++) {
          if (msgTools.filter(message, itemCb)) {
            continue;
          }
          itemCb(message.data, response);
        }
      });
    }
  };

  var chMessaging = {
    /**
     * @namespace chrome.runtime
     * @namespace chrome.tabs.query
     */
    cbList: [],
    currentTab: function (message) {
      chrome.tabs.query({active: true, currentWindow: true}, function(tabs) {
        if (tabs[0] === undefined || tabs[0].id < 0) {
          msgTools.rmCaller(message.monoResponseId);
          return;
        }
        message.monoTo = defaultId;
        chrome.tabs.sendMessage(tabs[0].id, message);
      });
    },
    send: function(message) {
      var tabId = message.tabId;
      if (tabId !== undefined && tabId > -1) {
        delete message.tabId;
        return chrome.tabs.sendMessage(tabId, message);
      }
      chrome.runtime.sendMessage(message);
    },
    on: function(cb) {
      msgTools.readFilter(this, cb);
      chMessaging.cbList.push(cb);
      if (chMessaging.cbList.length > 1) {
        return;
      }
      messagesEnable = true;
      chrome.runtime.onMessage.addListener(function(message, sender) {
        if (message.monoResponseId !== undefined) {
          return msgTools.cbCaller(message);
        }
        if (sender.tab !== undefined) {
          message.tabId = sender.tab.id;
        }
        var response = msgTools.mkResponse(message);
        for (var i = 0, itemCb; itemCb = chMessaging.cbList[i]; i++) {
          if (msgTools.filter(message, itemCb)) {
            continue;
          }
          itemCb(message.data, response);
        }
      });
    }
  };

  var opMessaging = {
    /**
     * @namespace opera.extension
     * @namespace opera.extension.tabs
     * @namespace opera.extension.tabs.getSelected
     * @namespace opera.extension.broadcastMessage
     */
    cbList: [],
    currentTab: function (message) {
      var currentTab = opera.extension.tabs.getSelected();
      message.monoTo = defaultId;
      try {
        currentTab.postMessage(message);
      } catch (e) {}
    },
    send: function(message) {
      var source = message.source;
      if (source !== undefined) {
        delete message.source;
        try {
          source.postMessage(message);
        } catch (e) {}
        return;
      }
      if (opera.extension.postMessage !== undefined) {
        return opera.extension.postMessage(message);
      }
      opera.extension.broadcastMessage(message);
    },
    on: function(cb) {
      msgTools.readFilter(this, cb);
      opMessaging.cbList.push(cb);
      if (opMessaging.cbList.length > 1) {
        return;
      }
      messagesEnable = true;
      opera.extension.onmessage = function(event) {
        var message = event.data;
        if (message.monoResponseId !== undefined) {
          return msgTools.cbCaller(message);
        }
        message.source = event.source;
        var response = msgTools.mkResponse(message);
        for (var i = 0, itemCb; itemCb = opMessaging.cbList[i]; i++) {
          if (msgTools.filter(message, itemCb)) {
            continue;
          }
          itemCb(message.data, response);
        }
      };
    }
  };

  var sfMessaging = {
    /**
     * @namespace safari.application
     * @namespace safari.application.activeBrowserWindow.activeTab
     * @namespace safari.extension.globalPage
     * @namespace safari.application.browserWindows
     */
    cbList: [],
    currentTab: function (message) {
      var currentTab = safari.application.activeBrowserWindow.activeTab;
      if (currentTab.page === undefined) {
        msgTools.rmCaller(message.monoResponseId);
        return;
      }
      message.monoTo = defaultId;
      currentTab.page.dispatchMessage("message", message);
    },
    send: function (message) {
      var source = message.source;
      if (source !== undefined) {
        delete message.source;
        if (source.page !== undefined) {
          return source.page.dispatchMessage("message", message);
        }
      }
      if (mono.isSafariPopup) {
        return safari.extension.globalPage.contentWindow.mono.safariDirectOnMessage({
          message: message,
          target: {page: {dispatchMessage: function(name, message) {
            mono.safariDirectOnMessage({message: message});
          }}}});
      }
      if (mono.isSafariBgPage) {
        for (var w = 0, window; window = safari.application.browserWindows[w]; w++) {
          for (var t = 0, tab; tab = window.tabs[t]; t++) {
            if (!tab.page || !tab.page.dispatchMessage) {
              continue;
            }
            tab.page.dispatchMessage("message", message);
          }
        }
        return;
      }
      safari.self.tab.dispatchMessage("message", message);
    },
    on: function (cb) {
      msgTools.readFilter(this, cb);
      sfMessaging.cbList.push(cb);
      if (sfMessaging.cbList.length > 1) {
        return;
      }
      messagesEnable = true;
      var onMessage = function(event) {
        var message = event.message;
        if (message.monoResponseId !== undefined) {
          return msgTools.cbCaller(message);
        }
        message.source = event.target;
        var response = msgTools.mkResponse(message);
        for (var i = 0, itemCb; itemCb = sfMessaging.cbList[i]; i++) {
          if (msgTools.filter(message, itemCb)) {
            continue;
          }
          itemCb(message.data, response);
        }
      };
      if ( (mono.isSafariPopup || mono.isSafariBgPage) && mono.safariDirectOnMessage === undefined ) {
        mono.safariDirectOnMessage = onMessage;
      }
      if (mono.isSafariBgPage) {
        return safari.application.addEventListener("message", onMessage, false);
      }
      safari.self.addEventListener("message", onMessage, false);
    }
  };

  var gmMessaging = {
    cbList: [],
    gotMsg: function() {},
    currentTab: function (message) {
      message.monoTo = defaultId;
      gmMessaging.gotMsg(message);
    },
    send: function (message) {
      gmMessaging.gotMsg(message);
    },
    on: function (cb) {
      msgTools.readFilter(this, cb);
      gmMessaging.cbList.push(cb);
      if (gmMessaging.cbList.length > 1) {
        return;
      }
      messagesEnable = true;
      gmMessaging.gotMsg = function (message) {
        if (message.monoResponseId !== undefined) {
          return msgTools.cbCaller(message);
        }
        var response = msgTools.mkResponse(message);
        for (var i = 0, itemCb; itemCb = gmMessaging.cbList[i]; i++) {
          if (msgTools.filter(message, itemCb)) {
            continue;
          }
          itemCb(message.data, response);
        }
      };
    }
  };

  mono.sendMessage = function(message, cb, to) {
    message = {
      data: message,
      monoTo: to || defaultId,
      monoFrom: mono.pageId
    };
    if (this.from !== undefined) {
      message.monoFrom = this.from;
    }
    if (this.timeout !== undefined) {
      message.timeout = this.timeout;
    }
    mono.debug.messages && console.log(mono.pageId, 'monoLog:', 'sendMessage', 'to:', to, message);
    msgTools.cbCollector(message, cb);
    if (to === 'activeTab') {
      return mono.sendMessage.currentTab(message);
    }
    mono.sendMessage.send(message);
  };

  if (mono.isChrome) {
    if (chrome.runtime && chrome.runtime.onMessage) {
      mono.sendMessage.send = chMessaging.send;
      mono.sendMessage.currentTab = chMessaging.currentTab;
      mono.onMessage = chMessaging.on;
    } else {
      mono.sendMessage.send = chRequestMessaging.send;
      mono.sendMessage.currentTab = chRequestMessaging.currentTab;
      mono.onMessage = chRequestMessaging.on;
    }
  } else
  if (mono.isFF) {
    if (mono.noAddon) {
      ffVirtualPort();
    }
    mono.sendMessage.send = ffMessaging.send;
    mono.sendMessage.currentTab = ffMessaging.currentTab;
    mono.onMessage = ffMessaging.on;
  } else
  if (mono.isOpera) {
    mono.sendMessage.send = opMessaging.send;
    mono.sendMessage.currentTab = opMessaging.currentTab;
    mono.onMessage = opMessaging.on;
  } else
  if (mono.isSafari) {
    mono.sendMessage.send = sfMessaging.send;
    mono.sendMessage.currentTab = sfMessaging.currentTab;
    mono.onMessage = sfMessaging.on;
  } else
  if (mono.isGM) {
    mono.sendMessage.send = gmMessaging.send;
    mono.sendMessage.currentTab = gmMessaging.currentTab;
    mono.onMessage = gmMessaging.on;
  }

  monoUtils(mono);

  return mono;
};

var monoUtils = function(mono) {
  if (mono.isModule) {
    XMLHttpRequest = require('sdk/net/xhr').XMLHttpRequest;
  }

  mono.parseXhrHeader = function(head) {
    head = head.replace(/\r?\n/g, '\n').split('\n');
    var obj = {
      monoParsed: 1
    };
    for (var i = 0, len = head.length; i < len; i++) {
      var keyValue = head[i].split(':');
      if (keyValue.length < 2) {
        continue;
      }
      var key = keyValue[0].trim();
      obj[key] = keyValue[1].trim();
    }
    return obj;
  };

  mono.ajax = function(obj) {
    var url = obj.url;

    var method = obj.type || 'GET';
    method.toUpperCase();

    var data = obj.data;

    if (data && typeof data !== "string") {
      data = mono.param(data);
    }

    if (data && method === 'GET') {
      url += ( (url.indexOf('?') === -1)?'?':'&' ) + data;
      data = undefined;
    }

    if (obj.cache === false && ['GET','HEAD'].indexOf(method) !== -1) {
      var nc = '_=' + Date.now();
      url += ( (url.indexOf('?') === -1)?'?':'&' ) + nc;
    }

    var xhr;
    if (mono.isGM) {
      xhr = {};
      xhr.method = method;
      xhr.url = url;
      xhr.data = data;
    } else {
      xhr = new XMLHttpRequest();
      xhr.open(method, url, true);
    }

    if (obj.dataType) {
      obj.dataType = obj.dataType.toLowerCase();

      if (!(mono.noXhrJson && obj.dataType === 'json')) {
        xhr.responseType = obj.dataType;
      }
    }

    if (!obj.headers) {
      obj.headers = {};
    }

    if (obj.contentType) {
      obj.headers["Content-Type"] = obj.contentType;
    }

    if (data && !obj.headers["Content-Type"]) {
      obj.headers["Content-Type"] = 'application/x-www-form-urlencoded; charset=UTF-8';
    }

    if (mono.isGM) {
      xhr.overrideMimeType = obj.mimeType;
      xhr.headers = obj.headers;
      xhr.onload = function(_xhr) {
        var xhr = mono.extend({}, _xhr);
        if (xhr.status >= 200 && xhr.status < 300 || xhr.status === 304) {
          var response = xhr.responseText;
          if (obj.dataType === 'json') {
            if (typeof xhr.response === 'string' && xhr.responseText) {
              try {
                response = JSON.parse(xhr.responseText);
              } catch (e) {}
            } else {
              response = xhr.response;
            }
          }
          if (obj.mimeType === 'text/xml') {
            var parser=new DOMParser();
            if (typeof xhr.response === 'string') {
              xhr.responseXML = parser.parseFromString(xhr.responseText, obj.mimeType);
            }
          }
          var responseHeaders = undefined;
          xhr.getResponseHeader = function(name) {
            if (responseHeaders === undefined) {
              responseHeaders = mono.parseXhrHeader(xhr.responseHeaders);
            }
            return responseHeaders[name];
          };
          return obj.success && obj.success(response, xhr);
        }
        obj.error && obj.error(xhr);
      };
      xhr.error = function(response) {
        obj.error && obj.error(response);
      };
      return GM_xmlhttpRequest(xhr);
    }

    if (obj.mimeType && !(mono.isOpera && obj.mimeType === 'application/json')) {
      xhr.overrideMimeType(obj.mimeType);
    }
    if (obj.headers) {
      for (var key in obj.headers) {
        xhr.setRequestHeader(key, obj.headers[key]);
      }
    }

    if (mono.isOpera || mono.isSafari) {
      xhr.onreadystatechange = function () {
        if (mono.isOpera && xhr.readyState > 1 && (xhr.status === 302 || xhr.status === 0)) {
          // Opera xhr redirect
          if (obj.noRedirect === undefined) {
            obj.noRedirect = 0;
          }
          var location = xhr.getResponseHeader('Location');
          if (location && obj.noRedirect < 5) {
            obj.noRedirect++;
            var _obj = mono.extend({}, obj);
            _obj.url = location;
            delete obj.success;
            delete obj.error;
            var _xhr = mono.ajax(_obj);
            xhr.abort = _xhr.abort;
          }
        }
        if (mono.isSafari && xhr.readyState > 1 && method === 'HEAD') {
          // Safari on HEAD 302 redirect fix
          obj.success && obj.success(undefined, xhr);
          delete obj.success;
          delete obj.error;
          xhr.abort();
        }
      };
    }

    xhr.onload = function () {
      if (xhr.status >= 200 && xhr.status < 300 || xhr.status === 304 ||
        ( (mono.isSafari || mono.isOpera) && xhr.status === 0 && xhr.response) ) {
        var response = (obj.dataType) ? xhr.response : xhr.responseText;
        if (obj.dataType === 'json' && typeof response !== 'object' && xhr.responseText) {
          try {
            response = JSON.parse(xhr.responseText);
          } catch (e) {}
        }
        return obj.success && obj.success(response, xhr);
      }
      obj.error && obj.error(xhr);
    };
    xhr.onerror = function(e) {
      obj.error && obj.error(xhr);
    };
    try {
      xhr.send(data);
    } catch (e) {
      // NS_ERROR_FILE_NOT_FOUND
      obj.error && obj.error({});
    }

    return xhr;
  };

  mono.extend = function() {
    var obj = arguments[0];
    for (var i = 1, len = arguments.length; i < len; i++) {
      var item = arguments[i];
      for (var key in item) {
        obj[key] = item[key];
      }
    }
    return obj;
  };

  mono.param = function(obj) {
    if (typeof obj === 'string') {
      return obj;
    }
    var itemsList = [];
    for (var key in obj) {
      if (!obj.hasOwnProperty(key)) {
        continue;
      }
      if (obj[key] === undefined) {
        obj[key] = '';
      }
      itemsList.push(encodeURIComponent(key)+'='+encodeURIComponent(obj[key]));
    }
    return itemsList.join('&');
  };

  mono.capitalize = function(word) {
    return word.charAt(0).toUpperCase() + word.substr(1);
  };

  mono.create = (function() {
    var createHook = {
      text: function(el, value) {
        el.textContent = value;
      },
      data: function(el, value) {
        for (var item in value) {
          el.dataset[item] = value[item];
        }
      },
      class: function(el, value) {
        if (typeof value !== 'string') {
          for (var i = 0, len = value.length; i < len; i++) {
            var className = value[i];
            el.classList.add(className);
          }
          return;
        }
        el.setAttribute('class', value);
      },
      style: function(el, value) {
        if (typeof value !== 'string') {
          for (var item in value) {
            el.style[item] = value[item];
          }
          return;
        }
        el.setAttribute('style', value);
      },
      append: function(el, value) {
        if (Array.isArray(value)) {
          for (var i = 0, len = value.length; i < len; i++) {
            var subEl = value[i];
            if (!subEl) {
              continue;
            }
            if (typeof (subEl) === 'string') {
              subEl = document.createTextNode(subEl);
            }
            el.appendChild(subEl);
          }
          return;
        }
        el.appendChild(value);
      },
      on: function(el, args) {
        if (typeof args[0] !== 'string') {
          for (var i = 0, len = args.length; i < len; i++) {
            var subArgs = args[i];
            mono.on(el, subArgs[0], subArgs[1], subArgs[2]);
          }
          return;
        }
        //type, onEvent, useCapture
        mono.on(el, args[0], args[1], args[2]);
      },
      onCreate: function(el, value) {
        value(el);
      }
    };

    return function(tagName, obj) {
      var el;
      if ( typeof tagName === 'string') {
        el = document.createElement(tagName);
      } else {
        el = tagName;
      }
      if (obj !== undefined) {
        for (var attr in obj) {
          var value = obj[attr];
          if (createHook[attr]) {
            createHook[attr](el, value);
            continue;
          }
          el[attr] = value;
        }
      }
      return el;
    }
  }());

  mono.trigger = function(el, type, data) {
    if (data === undefined) {
      data = {};
    }
    if (data.canBubble === undefined) {
      data.canBubble = false;
    }
    if (data.cancelable === undefined) {
      data.cancelable = false;
    }
    var event = new CustomEvent(type, data);
    el.dispatchEvent(event);
  };

  mono.str2regexp = function(s) {
    return new RegExp('^' + s.replace(/\./g, '\\.').replace(/\*/g, '.*?') + '$');
  };

  mono.checkUrl = function(url, rules) {
    return rules.some(function(rule){
      if (typeof rule === 'string') {
        rule = mono.str2regexp(rule);
      }
      return rule.test(url)
    });
  };

  mono.isIframe = function() {
    if (mono.isFF) {
      return window.parent !== window;
    } else {
      return window.top !== window.self;
    }
  };

  mono.uniFix = function() {
    if (mono.isOpera) {
      if (typeof location === 'undefined') {
        location = document.location;
      }
      if (typeof navigator === 'undefined') {
        navigator = window.navigator;
      }
      if (typeof localStorage === 'undefined') {
        localStorage = window.localStorage;
      }
      if (typeof CustomEvent === 'undefined') {
        CustomEvent = window.CustomEvent;
      }
      if (typeof XMLHttpRequest === 'undefined') {
        XMLHttpRequest = window.XMLHttpRequest;
      }
    }
    if (mono.isSafari && typeof CustomEvent === 'undefined') {
      CustomEvent = function (event, params) {
        params = params || { canBubble: false, cancelable: false };
        var evt = document.createEvent('CustomEvent');
        evt.initCustomEvent(event, params.canBubble, params.cancelable, params.detail);
        return evt;
      };
      CustomEvent.prototype = window.Event.prototype;
    }
  };

  mono.userJsCheck = function() {
    try {
      if(window.sessionStorage && window.sessionStorage['savefrom-helper-userjs'] === '1') {
        mono.sendMessage({action: 'userjsDetected'}, undefined, 'bg');
      }
    } catch (e) {}
  };

  mono.onReady = function(moduleName, cb) {
    if (mono.isGM) {
      _modules[moduleName] = function() { cb(moduleName); };
      return;
    }
    mono.uniFix();

    if (window.sessionStorage) {
      try {
        window.sessionStorage['savefrom-helper-extension'] = '1';
      } catch (e) {}
    }

    if (mono.isChrome || mono.isFF || document.readyState === 'complete') {
      return setTimeout(function() {
        cb(moduleName);
        mono.userJsCheck();
      });
    }

    var complete = function() {
      document.removeEventListener( "DOMContentLoaded", complete, false );
      window.removeEventListener( "load", complete, false );
      cb(moduleName);
    };
    document.addEventListener('DOMContentLoaded', complete, false);
    window.addEventListener('load', complete, false);
  };

  mono.openTab = function(url, select, active) {
    select = (select === undefined)?true:!!select;
    if (mono.isChrome) {
      var options = {url: url, selected: select};
      if (active) {
        options.active = !!active;
      }
      chrome.tabs.create(options);
    }
    if (mono.isFF) {
      var tabs = require("sdk/tabs");
      tabs.open(url);
    }
    if (mono.isSafari) {
      var window = safari.application.activeBrowserWindow;
      var tab = window.openTab();
      tab.url = url;
      if (select) {
        tab.activate();
      }
    }
    if (mono.isOpera) {
      opera.extension.tabs.create({ url: url, focused: select });
    }
    if (mono.isGM) {
      GM_openInTab(url, {
        active: select,
        insert: true
      });
    }
  };
  (function() {
    var newEvent = function (origType, _onEvent) {
      return function (e) {
        var event = {
          type: origType,
          target: e.target,
          stopPropagation: e.stopPropagation,
          preventDefault: e.preventDefault,
          stopImmediatePropagation: e.stopImmediatePropagation
        };
        _onEvent.call(this, event);
      };
    };

    var functionMap = {
      eventIndex: 0,
      replaceList: {}
    };

    mono.on = function(el, type, onEvent, useCapture) {
      if (type === 'mouseenter' || type === 'mouseleave') {
        if ((mono.isFF || (mono.isGM && !mono.isTM )) && el === document && !useCapture) {
          el = document.body;
        }

        if (mono.noMouseEnter) {
          var cacheEventKey = 'eventIndex_';
          var _onEvent = onEvent;

          var origType = type;
          if (type === 'mouseenter') {
            type = 'mouseover';
            cacheEventKey += 1;
          } else
          if (type === 'mouseleave') {
            type = 'mouseout';
            cacheEventKey += 0;
          }

          if (_onEvent[cacheEventKey] === undefined) {
            var eventIndex = functionMap.eventIndex;
            _onEvent[cacheEventKey] = eventIndex;

            onEvent = newEvent(origType, _onEvent);

            functionMap.replaceList[eventIndex] = onEvent;
            functionMap.eventIndex++;
          } else {
            onEvent = functionMap.replaceList[_onEvent[cacheEventKey]];
          }
        }
      }

      el.addEventListener(type, onEvent, useCapture);
    };

    mono.off = function(el, type, onEvent, useCapture) {
      if (type === 'mouseenter' || type === 'mouseleave') {
        if ((mono.isFF || (mono.isGM && !mono.isTM )) && el === document && !useCapture) {
          el = document.body;
        }

        if (mono.noMouseEnter) {
          var cacheEventKey = 'eventIndex_';
          if (type === 'mouseenter') {
            type = 'mouseover';
            cacheEventKey += 1;
          } else
          if (type === 'mouseleave') {
            type = 'mouseout';
            cacheEventKey += 0;
          }

          var eventIndex = onEvent[cacheEventKey];
          if (eventIndex !== undefined) {
            onEvent = functionMap.replaceList[eventIndex];
          }
        }
      }

      el.removeEventListener(type, onEvent, useCapture);
    };
  }());

  (function() {
    var vars = {
      lastUrl: undefined,
      timer: undefined,
      enable: false,
      eventList: []
    };

    var onUrlChangeListener = function() {
      var url = document.location.href;
      var lastUrl = vars.lastUrl;
      if (lastUrl === url) {
        return;
      }
      vars.lastUrl = url;

      for (var i = 0, len = vars.eventList.length; i < len; i++) {
        vars.eventList[i](url, lastUrl);
      }
    };

    mono.onUrlChange = function(cb, now) {
      if (vars.eventList.indexOf(cb) !== -1) {
        return;
      }
      var currentUrl = window.location.href;

      vars.eventList.push(cb);

      now && cb(currentUrl);

      if (vars.enable) {
        return;
      }
      vars.enable = true;
      vars.lastUrl = currentUrl;

      vars.timer = setInterval(function() {
        onUrlChangeListener();
      }, 1000);

      // window.addEventListener('popstate', onUrlChangeListener);
    };

    mono.offUrlChange = function(cb) {
      var urlChange = vars;
      var pos = urlChange.eventList.indexOf(cb);
      if (pos === -1) {
        return;
      }
      urlChange.eventList.splice(pos, 1);
      if (urlChange.eventList.length === 0) {
        clearInterval(vars.timer);
        // window.removeEventListener('popstate', onUrlChangeListener);
      }
    };
  }());

  mono.global = {};

  mono.getParentByClass = function(el, classList) {
    if (typeof classList === 'string') {
      classList = [classList];
    }

    for(var parent = el; parent; parent = parent.parentNode) {
      if (parent.nodeType !== 1) {
        return null;
      }
      for (var i = 0, className; className = classList[i]; i++) {
        if (parent.classList.contains(className)) {
          return parent;
        }
      }
    }

    return null;
  };

  mono.parseUrlParams = function(url) {
    var startFrom = url.indexOf('?');
    var query = url;
    if (startFrom !== -1) {
      query = url.substr(startFrom + 1);
    }
    var dblParamList = query.split('&');
    var params = {};
    for (var i = 0, len = dblParamList.length; i < len; i++) {
      var item = dblParamList[i];
      var ab = item.split('=');
      params[ab[0]] = ab[1] || '';
    }
    return params;
  };

  // legacy

  mono.getQueryString = function(query, key_prefix, key_suffix) {
    if(!query || typeof(query) != 'object')
      return '';

    if(key_prefix === undefined)
      key_prefix = '';

    if(key_suffix === undefined)
      key_suffix = '';

    var str = '';
    for(key in query)
    {
      if(str.length)
        str += '&';

      if(query[key] instanceof Object)
      {
        if(!key_prefix)
          key_prefix = '';

        if(!key_suffix)
          key_suffix = '';

        str += getQueryString(query[key], key_prefix + key + "[", "]" + key_suffix);
      }
      else
        str += key_prefix + escape(key) + key_suffix + '=' + escape(query[key]);
    }

    return str;
  };

  mono.decodeUnicodeEscapeSequence = function(text) {
    return text.replace(/\\u([0-9a-f]{4})/g, function(s, m){
      m = parseInt(m, 16);
      if(!isNaN(m))
      {
        return String.fromCharCode(m);
      }
    });
  };

  mono.decodeHTML = function(html) {
    var d = document.createElement('div');
    d.innerHTML = html.replace(/</g, '&lt;').
      replace(/\>/g, '&gt;').
      replace(/\"/g, '&quot;');

    var res = d.textContent;
    d.textContent = '';
    return mono.decodeUnicodeEscapeSequence(res);
  };

  mono.fileName = {
    maxLength: 80,

    getParts: function (name) {
      return name.match(/^(.+)\.([a-z0-9]{1,4})$/i);
    },

    shorten: function (name) {
      name = name.substr(0, this.maxLength);
      var nws = '\\u0030-\\u0039\\u0041-\\u005a\\u005f\\u0061-\\u007a\\u00aa\\u00b5\\u00ba\\u00c0-\\u00d6\\u00d8-\\u00f6\\u00f8-\\u021f\\u0222-\\u0233\\u0250-\\u02ad\\u0386\\u0388-\\u038a\\u038c\\u038e-\\u03a1\\u03a3-\\u03ce\\u03d0-\\u03d7\\u03da-\\u03f3\\u0400-\\u0481\\u048c-\\u04c4\\u04c7\\u04c8\\u04cb\\u04cc\\u04d0-\\u04f5\\u04f8\\u04f9\\u0531-\\u0556\\u0561-\\u0587\\u05d0-\\u05ea\\u05f0-\\u05f2\\u0621-\\u063a\\u0641-\\u064a\\u0660-\\u0669\\u0671-\\u06d3\\u06d5\\u06f0-\\u06fc\\u0710\\u0712-\\u072c\\u0780-\\u07a5\\u0905-\\u0939\\u093d\\u0950\\u0958-\\u0961\\u0966-\\u096f\\u0985-\\u098c\\u098f\\u0990\\u0993-\\u09a8\\u09aa-\\u09b0\\u09b2\\u09b6-\\u09b9\\u09dc\\u09dd\\u09df-\\u09e1\\u09e6-\\u09f1\\u0a05-\\u0a0a\\u0a0f\\u0a10\\u0a13-\\u0a28\\u0a2a-\\u0a30\\u0a32\\u0a33\\u0a35\\u0a36\\u0a38\\u0a39\\u0a59-\\u0a5c\\u0a5e\\u0a66-\\u0a6f\\u0a72-\\u0a74\\u0a85-\\u0a8b\\u0a8d\\u0a8f-\\u0a91\\u0a93-\\u0aa8\\u0aaa-\\u0ab0\\u0ab2\\u0ab3\\u0ab5-\\u0ab9\\u0abd\\u0ad0\\u0ae0\\u0ae6-\\u0aef\\u0b05-\\u0b0c\\u0b0f\\u0b10\\u0b13-\\u0b28\\u0b2a-\\u0b30\\u0b32\\u0b33\\u0b36-\\u0b39\\u0b3d\\u0b5c\\u0b5d\\u0b5f-\\u0b61\\u0b66-\\u0b6f\\u0b85-\\u0b8a\\u0b8e-\\u0b90\\u0b92-\\u0b95\\u0b99\\u0b9a\\u0b9c\\u0b9e\\u0b9f\\u0ba3\\u0ba4\\u0ba8-\\u0baa\\u0bae-\\u0bb5\\u0bb7-\\u0bb9\\u0be7-\\u0bef\\u0c05-\\u0c0c\\u0c0e-\\u0c10\\u0c12-\\u0c28\\u0c2a-\\u0c33\\u0c35-\\u0c39\\u0c60\\u0c61\\u0c66-\\u0c6f\\u0c85-\\u0c8c\\u0c8e-\\u0c90\\u0c92-\\u0ca8\\u0caa-\\u0cb3\\u0cb5-\\u0cb9\\u0cde\\u0ce0\\u0ce1\\u0ce6-\\u0cef\\u0d05-\\u0d0c\\u0d0e-\\u0d10\\u0d12-\\u0d28\\u0d2a-\\u0d39\\u0d60\\u0d61\\u0d66-\\u0d6f\\u0d85-\\u0d96\\u0d9a-\\u0db1\\u0db3-\\u0dbb\\u0dbd\\u0dc0-\\u0dc6\\u0e01-\\u0e30\\u0e32\\u0e33\\u0e40-\\u0e45\\u0e50-\\u0e59\\u0e81\\u0e82\\u0e84\\u0e87\\u0e88\\u0e8a\\u0e8d\\u0e94-\\u0e97\\u0e99-\\u0e9f\\u0ea1-\\u0ea3\\u0ea5\\u0ea7\\u0eaa\\u0eab\\u0ead-\\u0eb0\\u0eb2\\u0eb3\\u0ebd\\u0ec0-\\u0ec4\\u0ed0-\\u0ed9\\u0edc\\u0edd\\u0f00\\u0f20-\\u0f29\\u0f40-\\u0f47\\u0f49-\\u0f6a\\u0f88-\\u0f8b\\u1000-\\u1021\\u1023-\\u1027\\u1029\\u102a\\u1040-\\u1049\\u1050-\\u1055\\u10a0-\\u10c5\\u10d0-\\u10f6\\u1100-\\u1159\\u115f-\\u11a2\\u11a8-\\u11f9\\u1200-\\u1206\\u1208-\\u1246\\u1248\\u124a-\\u124d\\u1250-\\u1256\\u1258\\u125a-\\u125d\\u1260-\\u1286\\u1288\\u128a-\\u128d\\u1290-\\u12ae\\u12b0\\u12b2-\\u12b5\\u12b8-\\u12be\\u12c0\\u12c2-\\u12c5\\u12c8-\\u12ce\\u12d0-\\u12d6\\u12d8-\\u12ee\\u12f0-\\u130e\\u1310\\u1312-\\u1315\\u1318-\\u131e\\u1320-\\u1346\\u1348-\\u135a\\u1369-\\u1371\\u13a0-\\u13f4\\u1401-\\u166c\\u166f-\\u1676\\u1681-\\u169a\\u16a0-\\u16ea\\u1780-\\u17b3\\u17e0-\\u17e9\\u1810-\\u1819\\u1820-\\u1842\\u1844-\\u1877\\u1880-\\u18a8\\u1e00-\\u1e9b\\u1ea0-\\u1ef9\\u1f00-\\u1f15\\u1f18-\\u1f1d\\u1f20-\\u1f45\\u1f48-\\u1f4d\\u1f50-\\u1f57\\u1f59\\u1f5b\\u1f5d\\u1f5f-\\u1f7d\\u1f80-\\u1fb4\\u1fb6-\\u1fbc\\u1fbe\\u1fc2-\\u1fc4\\u1fc6-\\u1fcc\\u1fd0-\\u1fd3\\u1fd6-\\u1fdb\\u1fe0-\\u1fec\\u1ff2-\\u1ff4\\u1ff6-\\u1ffc\\u203f\\u2040\\u207f\\u2102\\u2107\\u210a-\\u2113\\u2115\\u2119-\\u211d\\u2124\\u2126\\u2128\\u212a-\\u212d\\u212f-\\u2131\\u2133-\\u2139\\u3006\\u3041-\\u3094\\u30a1-\\u30fb\\u3105-\\u312c\\u3131-\\u318e\\u31a0-\\u31b7\\u3400-\\u4db5\\u4e00-\\u9fa5\\ua000-\\ua48c\\uac00-\\ud7a3\\uf900-\\ufa2d\\ufb00-\\ufb06\\ufb13-\\ufb17\\ufb1d\\ufb1f-\\ufb28\\ufb2a-\\ufb36\\ufb38-\\ufb3c\\ufb3e\\ufb40\\ufb41\\ufb43\\ufb44\\ufb46-\\ufbb1\\ufbd3-\\ufd3d\\ufd50-\\ufd8f\\ufd92-\\ufdc7\\ufdf0-\\ufdfb\\ufe33\\ufe34\\ufe4d-\\ufe4f\\ufe70-\\ufe72\\ufe74\\ufe76-\\ufefc\\uff10-\\uff19\\uff21-\\uff3a\\uff3f\\uff41-\\uff5a\\uff65-\\uff6f\\uff71-\\uff9d\\uffa0-\\uffbe\\uffc2-\\uffc7\\uffca-\\uffcf\\uffd2-\\uffd7\\uffda-\\uffdc';
      var re = new RegExp('^(.+)([^' + nws + ']+)[' + nws + ']+$', 'i');
      return name.replace(re, function (str, m1, m2) {
        m1 = m1.replace(/[-,_\.]+$/g, '');
        if (m2 == ')' || m2 == ']')
          return m1 + m2;

        return m1;
      });
    },

    modify: function (name) {
      if (!name)
        return '';

      name = name.replace(/^\s+|\s+$/g, '');

      try {
        name = decodeURIComponent(name);
      }
      catch (err) {
        name = unescape(name);
      }

      name = mono.decodeHTML(name);

      name = name.replace(/\s+/g, ' ')
        .replace(/[\x2F\x5C\x3A\x7C]+/g, '_')
        .replace(/[\x2A\x3F\x22]/g, '')
        .replace(/\x3C/g, '(')
        .replace(/\x3E/g, ')')
        .replace(/(?:^\s+)|(?:\s+$)/g, '')
        .replace(/(\.|\!|\?|_|,|\-|\:|\+){2,1000}/g, '$1')
        .replace(/[\.,:;\/\-_\+=\x27]$/ig, '');

      if (name[0] === '.') {
        name = name.substr(1);
      }

      if (!name)
        return '';

      if (name.length <= this.maxLength)
        return name;

      var parts = this.getParts(name);
      if (parts && parts.length == 3) {
        parts[1] = this.shorten(parts[1]);
        return parts[1] + '.' + parts[2];
      }

      name = this.shorten(name);
      return name;
    }
  };
};

if (typeof window !== "undefined") {
  mono = mono(window);
} else {
  setTimeout = require("sdk/timers").setTimeout;
  exports.init = mono;
}