// ==UserScript==
// @name        YouTube downloader
//
// @include     http://youtube.com/*
// @include     http://*.youtube.com/*
// @include     https://youtube.com/*
// @include     https://*.youtube.com/*
// ==/UserScript==

mono.onReady('youtube', function(moduleName) {
  if (mono.isSafari) {
    if (!mono.checkUrl(document.URL, [
      'http://youtube.com/*',
      'http://*.youtube.com/*',
      'https://youtube.com/*',
      'https://*.youtube.com/*'
    ])) {
      return;
    }
  }

  var language = {};
  var preference = {};
  var init = function () {
    mono.pageId = moduleName;
    mono.onMessage.call({filter: moduleName}, function(message, cb){
      if (!preference.enable || !preference.moduleYoutube) {
        return;
      }
      if (message.action === 'getPageId') {
        return cb(mono.pageId);
      }
    });
    mono.sendMessage(['getLanguage', 'getPreference'], function (response) {
      language = response.getLanguage;
      preference = response.getPreference;
      if (!preference.enable || !preference.moduleYoutube) {
        return;
      }
      mono.global = {
        language: language,
        preference: preference
      };
      youtube.run();
    }, 'bg');
  };

  var iframe = mono.isIframe();

  if(iframe) {
    var m = location.href.match(/\/embed\/([\w\-]+)/i);

    if(!m || m.length < 2) {
      return;
    }
  }

  var youtube = {
    swfargs: null,
    token: '',
    video_id: '',
    sts: '',
    needDecipher: false,
    newInterface: false,

    panelId: 'savefrom__yt_links',
    buttonId: 'savefrom__yt_btn',
    panelParent: null,
    panelInsertBefore: null,
    timer: null,

    btnBox: null,
    links: {},
    isMobile: false,

    run: function(vid, links)
    {
      if (location.host.indexOf('m.') === 0) {
        youtube.isMobile = true;
        mono.onUrlChange(function(url) {
          SaveFrom_Utils.mobileLightBox.clear();
          var vid = youtube.getIdFromLocation(url);
          if (!vid) {
            youtube.waitBtnContainer.stop();
            return;
          }
          youtube.waitBtnContainer.start(function() {
            return youtube.getMobileContainer();
          }, function(container) {
            youtube.appendMobileButton(vid, container);
          });
        }, 1);
        return;
      }

      if (vid && links) {
        youtube.links[vid] = links;
      }

      if(iframe) {
        youtube.isVideoPage();
        youtube.video_id = youtube.swfargs.video_id;
        youtube.appendIframeButtons();
        return;
      }

      if(youtube.isVideoPage() || links)
      {
        if(vid) {
          youtube.video_id = vid;
        }

        var p = youtube.findPanel();
        if(p) {
          youtube.appendDownloadButton(p);
        }
      }

      if(!iframe)
      {
        window.clearInterval(youtube.timer);

        var loc = document.location.href;
        youtube.timer = window.setInterval(function(){
          if(document.location.href != loc)
          {
            loc = document.location.href;

            var e = document.getElementById(youtube.panelId);
            if (e !== null && e.style.display !=='none') {
              youtube.closePopup();
            }

            var vid = youtube.getIdFromLocation();
            if(!vid)
              return;

            mono.sendMessage({action: 'getYoutubeLinks', extVideoId: vid, url: location.href}, function(response){
              if(response.links) {
                youtube.links[vid] = response.links;
                youtube.catchVideoChange(loc, vid, response.links);
              }
            }, 'bg');
          }
        }, 1000);
      }
    },

    waitBtnContainer: {
      count: undefined,
      timer: undefined,
      check: function(getContainer, onReady) {
        var container = getContainer();
        if (container !== null && container.dataset.sfSkip === undefined) {
          if (container.dataset.sfFound !== undefined) {
            container.dataset.sfSkip = 1;
          } else {
            if (container.dataset.sfCondidate === undefined) {
              container.dataset.sfCondidate = 1;
            } else {
              container.dataset.sfFound = 1;
              onReady(container);
              return 1;
            }
          }
        }

        return undefined;
      },
      start: function(getContainer, onReady) {
        var _this = youtube.waitBtnContainer;

        _this.stop();

        if (_this.check(getContainer, onReady) !== undefined) {
          return;
        }

        _this.timer = setInterval(function() {
          _this.count--;
          if (_this.check(getContainer, onReady) !== undefined || _this.count <= 0) {
            clearInterval(_this.timer);
          }
        }, 250);
      },
      stop: function() {
        var _this = youtube.waitBtnContainer;

        clearInterval(_this.timer);
        _this.count = 20;
      }
    },

    catchVideoChange: function(loc, vid, links)
    {
      var count = 0;

      var timer = setInterval(function() {
        if(loc != document.location.href)
        {
          window.clearInterval(timer);
          return;
        }

        if(!document.getElementById(youtube.buttonId) && !!youtube.findPanel())
        {
          window.clearInterval(timer);
          youtube.run(vid, links);
        }

        if(count > 20)
          window.clearInterval(timer);

        count++;
      }, 1000);
    },


    findPanel: function()
    {
      var p = document.getElementById('watch-headline-user-info');
      if(!p)
      {
        p = document.getElementById('watch7-subscription-container');
        if(p)
          youtube.newInterface = true;
      }

      if(!p)
      {
        var e = document.getElementById('watch-video-count');
        if(!e)
          e = document.getElementById('editSubscriptionDiv');

        if(e)
          p = e.parentNode;
      }

      if(p)
      {
        youtube.panelParent = p.parentNode;
        youtube.panelInsertBefore = SaveFrom_Utils.nextSibling(p);
        //youtube.appendDownloadButton(p);
      }

      return p;
    },


    getIdFromLocation: function(loc)
    {
      if(!loc)
        loc = document.location.href;

      var m = loc.match(/\/watch\?(?:.+&)?v=([\w\-]+)/i);
      if(m && m.length > 1)
        return m[1];

      return null;
    },


    isVideoPage: function()
    {
      var swfargs = youtube.getSwfArgs();
      if(!swfargs)
        return false;

      youtube.swfargs = swfargs;

      var token = swfargs.t ? swfargs.t : swfargs.token;
      if(!token)
        return false;

      youtube.token = token;

      var video_id = swfargs.video_id;
      if(!video_id)
        video_id = youtube.getIdFromLocation();

      if(!video_id)
        return false;

      youtube.video_id = video_id;

      youtube.sts = SaveFrom_Utils.getMatchFirst(document.body.innerHTML,
        /[\"']sts[\"']\s*:\s*[\"']?(\d+)/i);

      return true;
    },

    getSwfArgs: function()
    {
      var html = document.body.innerHTML;

      var t = html.match(/[\"\'](?:swf_)?args[\"\']\s*:\s*(\{[^\}]+\})/i);
      if(t && t.length > 1)
      {
        if(t[1].search(/:\s*\"[^\"]+\s*\}$/) > -1)
          t[1] = t[1].replace(/\s*\}$/i, '"}');

        try
        {
          return JSON.parse(t[1]);
        }
        catch(err){
          t = html.match(/ytplayer.config\s*=\s*(\{.*?\});?<\/script>/i);
          if(t && t.length > 1) {
            try
            {
              return JSON.parse(t[1]);
            }
            catch(err){}
          }
        }
      }
      else
      {
        t = html.match(/flashvars\s*=\s*\\?[\"']?([^\"']+)/i);
        if(t && t.length > 1)
        {
          t = t[1];
          if(t.indexOf(/&/) == -1)
            t = t.replace(/(?:\\u0026|%26|%2526|&)amp;/ig, '&');

          t = mono.parseUrlParams(t);
          for(var i in t)
            t[i] = decodeURIComponent(t[i]);

          return t;
        }
      }

      var e = document.getElementsByTagName('embed');
      for(var i = 0; i < e.length; i++)
      {
        var f = e[i].getAttribute('flashvars');
        if(f && f.search(/fmt_map=/i) != -1)
        {
          return mono.parseUrlParams(f);
        }
      }

      return null;
    },

    getMobileContainer: function() {
      var elList = document.querySelectorAll('a[onclick][href="#"] span[id]');
      var elCount = 0;
      var fEl = undefined;
      for (var i = 0, el; el=elList[i];i++) {
        if (!el.id || el.id.substr(0, 10) !== 'koya_elem_') {
          continue;
        }
        elCount++;
        fEl = el;
      }
      if (elCount < 3) {
        return null;
      }
      var parent = fEl.parentNode.parentNode.parentNode;
      if (parent === null) {
        return null;
      }
      return parent.parentNode;
    },

    appendMobileButton: function(vid, container) {
      container.appendChild(mono.create('div', {
        data: {
          id: vid
        },
        style: {
          display: 'inline-block',
          height: '28px',
          width: '18px',
          marginRight: '20px',
          background: 'url('+SaveFrom_Utils.svg.getSrc('download', '#ADADAD')+') center no-repeat',
          cssFloat: 'right'
        },
        on: ['click', function() {
          var title = youtube.getTitle();
          if(title) {
            title = modifyTitle(title);
          }

          var vid = this.dataset.id;

          if (youtube.links[vid]) {
            var menuLinks = SaveFrom_Utils.popupMenu.prepareLinks.youtube(youtube.links[vid], title);
            return SaveFrom_Utils.mobileLightBox.show(menuLinks);
          }

          var request = {
            action: 'getYoutubeLinks',
            extVideoId: vid,
            url: location.href
          };

          var LightBox = SaveFrom_Utils.mobileLightBox.show(language.download + ' ...');

          mono.sendMessage(request, function(response){
            var menuLinks = undefined;
            if (response.links) {
              youtube.links[vid] = response.links;
              menuLinks = SaveFrom_Utils.popupMenu.prepareLinks.youtube(response.links, title);
            }
            if (!menuLinks) {
              menuLinks = language.noLinksFound;
            }
            LightBox.update(menuLinks);
          }, 'bg');

          LightBox.update(language.download + ' ...');
        }]
      }))
    },

    appendDownloadButton: function(parent)
    {
      if(document.getElementById(youtube.buttonId)) {
        return;
      }

      // var b = document.createElement('button');
      // b.id = youtube.buttonId;
      var selectBtn = undefined;
      var buttonContainer = mono.create('div', {
        id: youtube.buttonId,
        style: {
          display: 'inline-block',
          marginLeft: '10px',
          verticalAlign: 'middle'
        },
        append: [
          mono.create('a', {
            class: 'sf-quick-dl-btn',
            style: {
              display: 'inline-block',
              fontSize: 'inherit',
              height: '22px',
              border: '1px solid #00B75A',
              borderRadius: '3px',
              borderTopRightRadius: 0,
              borderBottomRightRadius: 0,
              paddingRight: '12px',
              paddingLeft: '28px',
              color: '#fff',
              cursor: 'pointer',
              verticalAlign: 'middle',
              position: 'relative',
              href: '#',
              lineHeight: '22px',
              textDecoration: 'none',
              zIndex: 1
            },
            append: [
              mono.create('i', {
                style: {
                  position: 'absolute',
                  display: 'inline-block',
                  left: '6px',
                  top: '3px',
                  backgroundImage: 'url('+SaveFrom_Utils.svg.getSrc('download', '#ffffff')+')',
                  backgroundSize: '12px',
                  backgroundRepeat: 'no-repeat',
                  backgroundPosition: 'center',
                  width: '16px',
                  height: '16px'
                }
              }),
              language.download
            ]
          }),
          mono.create('style', {text: '' +
            '#' + youtube.buttonId + ' .sf-quick-dl-btn {' +
            'background-color: #00B75A;' +
            '}' +
            '#' + youtube.buttonId + ' .sf-quick-dl-btn:hover {' +
            'background-color: rgb(0, 163, 80);' +
            '}' +
            '#' + youtube.buttonId + ' .sf-quick-dl-btn:active {' +
            'background-color: rgb(0, 151, 74);' +
            '}' +
          ''}),
          selectBtn = mono.create('button', {
            style: {
              position: 'relative',
              display: 'inline-block',
              marginLeft: '-2px',
              fontSize: 'inherit',
              height: '24px',
              paddingRight: '21px',
              backgroundColor: '#F8F8F8',
              border: '1px solid #CCCCCC',
              borderRadius: '3px',
              borderTopLeftRadius: '0',
              borderBottomLeftRadius: '0',
              cursor: 'pointer',
              color: '#9B9B9B',
              zIndex: 0,
              verticalAlign: 'middle'
            },
            on: ['mousedown', youtube.getLinks],
            append: [
              mono.create('i', {
                style: {
                  position: 'absolute',
                  display: 'inline-block',
                  top: '9px',
                  right: '6px',
                  border: '5px solid #868282',
                  borderBottomColor: 'transparent',
                  borderLeftColor: 'transparent',
                  borderRightColor: 'transparent'
                }
              })
            ]
          })
        ]
      });
      youtube.setButtonValue(selectBtn);

      parent.appendChild(buttonContainer);
    },


    updateWmode: function(p)
    {
      var e = p.querySelector('embed');
      if(!e) // html5
        return;

      // check wmode
      var m = e.getAttribute('wmode');
      if(m != 'opaque' && m != 'transparent')
      {
        // update wmode
        e.setAttribute('wmode', 'opaque');
        p.removeChild(e);
        p.appendChild(e);
      }
    },

    appendIframeButtons: function()
    {
      var p = document.getElementById('player'),
        b = document.createElement('div'),
        a = document.createElement('a'),
        panel = document.createElement('div');

      setTimeout(function() {
        youtube.updateWmode(p);
      }, 50);

      a.href = '#';
      a.textContent = language.download.toLowerCase();
      SaveFrom_Utils.setStyle(a, {
        display: 'inline-block',
        color: 'rgba(255,255,255,.9)',
        textDecoration: 'none',
        padding: '5px 10px'
      });
      b.appendChild(a);

      SaveFrom_Utils.setStyle(b, {
        background: 'rgba(0, 0, 0, .4)',
        border: '1px solid rgba(255,255,255,.5)',
        borderRadius: '4px',
        fontFamily: 'Arial,Helvetica,sans-serif',
        fontSize: '13px',
        lineHeight: 'normal',
        position: 'absolute',
        top: '30px',
        right: '5px',
        padding: 0,
        margin: 0,
        zIndex: 99999
      });

      panel.id = youtube.panelId;
      SaveFrom_Utils.setStyle(panel, {
        color: '#fff',
        background: 'rgba(0,0,0,0.7)',
        textAlign: 'center',
        border: 0,
        display: 'none',
        fontFamily: 'Arial,Helvetica,sans-serif',
        fontSize: '13px',
        fontWeight: 'normal',
        lineHeight: 'normal',
        position: 'absolute',
        top: '25px',
        left: 0,
        right: 0,
        margin: 0,
        padding: '3px',
        zIndex: 99990
      });

      if(document.body.scrollWidth <= 400)
      {
        panel.style.paddingTop = '28px';
      }

      youtube.btnBox = document.createElement('div');
      youtube.btnBox.style.display = 'none';
      youtube.btnBox.appendChild(b);
      youtube.btnBox.appendChild(panel);

      mono.off(document, 'mouseenter', youtube.onExtPlayerOver, false);
      mono.off(document, 'mouseleave', youtube.onExtPlayerOver, false);
      mono.on(document, 'mouseenter', youtube.onExtPlayerOver, false);
      mono.on(document, 'mouseleave', youtube.onExtPlayerOver, false);

      a.addEventListener('click', youtube.fetchIframeLinks, false);
      a.addEventListener('click', youtube.toggleIframePanel, false);

      document.body.appendChild(youtube.btnBox);
    },


    fetchIframeLinks: function(e)
    {
      e.preventDefault();
      e.stopPropagation();

      var button = e.target;

      var vid = youtube.video_id;

      var request = {
        action: 'getYoutubeLinks',
        extVideoId: vid,
        url: location.href
      };

      mono.sendMessage(request, function(response){
        if(response.links)
        {
          youtube.links[vid] = response.links;
        }
        youtube.appendIframeLinks(response.links, button);
      }, 'bg');

      youtube.appendIframeLinks(language.download + ' ...', button);
    },


    appendIframeLinks: function(links, button)
    {
      var panel = document.getElementById(youtube.panelId);
      if (!links || links.length === 0) {
        links = language.noLinksFound;
      }
      if(typeof(links) == 'object')
      {
        var title = youtube.getTitle(), titleAttr = '';
        if(title)
        {
          title = modifyTitle(title);
          /*
          titleAttr = '&title=' + encodeURIComponent(title);
          for(var i in links) {
            links[i] += titleAttr;
          }
          */
        }

        panel.textContent = '';
        SaveFrom_Utils.video.yt.init();
        SaveFrom_Utils.video.yt.show(links, panel, preference.moduleShowDownloadInfo, {
          link: {color: '#fff', borderBottom: '1px solid #808080',
            whiteSpace: 'nowrap', textDecoration: 'none'},
          text: {color: '#d0d0d0'},
          btn: {color: '#d0d0d0'},
          fsIcon: {color: '#ffffff', opacity: '.5'},
          fsText: {color: '#d0d0d0'}
        }, title);

        button.removeEventListener('click', youtube.fetchIframeLinks, false);
      }
      else if(typeof(links) == 'string')
      {
        panel.innerHTML = links;
      }
    },


    toggleIframePanel: function(e)
    {
      e.preventDefault();
      e.stopPropagation();

      var panel = document.getElementById(youtube.panelId);
      if(panel)
      {
        panel.style.display = (panel.style.display == 'none') ? '' : 'none';
      }
    },


    onExtPlayerOver: function(event)
    {
      if(youtube.btnBox)
      {
        if(event.type == 'mouseenter')
        {
          if(youtube.btnBox.style.display == 'none')
            youtube.btnBox.style.display = 'block';
        }
        else if(event.type == 'mouseleave')
        {
          youtube.btnBox.style.display = 'none';
        }
      }
    },


    getTitle: function()
    {
      var t = document.getElementById('watch-headline-title');
      if(t)
        return t.textContent;

      var meta = document.getElementsByTagName('meta');
      for(var i = 0; i < meta.length; i++)
      {
        var name = meta[i].getAttribute('name');
        if(name && name.toLowerCase() == 'title')
          return meta[i].getAttribute('content');
      }

      if(iframe || youtube.isMobile)
        return document.title.replace(/ - YouTube$/, '');

      return '';
    },

    onPopupMenuItemClick: function(button, data) {
      var isBest = data.isBest ? 'yes' : 'no';

      var quality = data.quality || '';
      if (quality) {
        quality = quality.replace(' ' + mono.global.language.kbps, '');
      }
      var label = [
        data.format, quality
      ];
      if (data['3d']) {
        label.push(data['3d']);
      }
      if (data.noAudio) {
        label.push(data.noAudio);
      }
      label.push(data.itag);
      if (data.quality === 'ummy') {
        label = [data.format];
      }

      label = label.join(' ');

      mono.sendMessage({action: 'trackEvent', category: 'youtube', event: 'download', label: label}, undefined, 'bg');
      mono.sendMessage({action: 'trackEvent', category: 'youtube', event: 'download_best', label: isBest}, undefined, 'bg');

      if (data.format === '???') {
        mono.sendMessage({action: 'trackEvent', category: 'youtube', event: 'new_format', label: data.itag+' '+youtube.video_id}, undefined, 'bg');
      }

      var itag = data.itag;
      mono.storage.set({ytLastITag: itag}, function() {
        if (data.noUpdate) {
          return;
        }
        youtube.setButtonValue(button);
      });
    },

    showLinks: function(links, button)
    {
      var _button = document.getElementById(youtube.buttonId);
      // _button = is container,
      // button - select btn

      if (typeof links === 'string') {
        SaveFrom_Utils.popupMenu.create({
          links: links,
          button: _button || button,
          popupId: youtube.panelId
        });
      } else {
        var title = youtube.getTitle();
        if(title) {
          title = modifyTitle(title);
        }

        var menuLinks = SaveFrom_Utils.popupMenu.prepareLinks.youtube(links, title, SaveFrom_Utils);

        SaveFrom_Utils.popupMenu.create({
          links: menuLinks,
          button: _button || button,
          popupId: youtube.panelId,
          showFileSize: true,
          onItemClick: function(data) {
            youtube.onPopupMenuItemClick(button, data);
          }
        });
      }

      // button.classList.add('yt-uix-button-toggled');

      button.removeEventListener('mousedown', youtube.getLinks, false);
      button.removeEventListener('mousedown', youtube.togglePanel, false);
      button.addEventListener('mousedown', youtube.togglePanel, false);

      document.removeEventListener('mousedown', youtube.closePopup, false);
      document.addEventListener('mousedown', youtube.closePopup, false);
    },

    closePopup: function () {
      youtube.togglePanel();
      document.removeEventListener('mousedown', youtube.closePopup);
    },

    togglePanel: function(event)
    {
      var button = document.getElementById(youtube.buttonId);

      if (event) {
        event.preventDefault();
        event.stopPropagation();
      }

      var e = document.getElementById(youtube.panelId);
      if(e)
      {
        if(e.style.display == 'none')
        {
          e.style.display = '';

          document.addEventListener('mousedown', youtube.closePopup, false);
        }
        else
        {
          e.style.display = 'none';

          document.removeEventListener('mousedown', youtube.closePopup, false);
        }
      }
    },

    onDlBtnClick: function(e, listItem, button) {
      if (!listItem) {
        e.preventDefault();
        e.stopPropagation();
        mono.trigger(this.parentNode.lastChild, 'mousedown');
        return;
      }

      var data = {
        itag: listItem.itag || '',
        quality: listItem.quality || '',
        format: listItem.format || '???',
        '3d': listItem['3d'] ? '3D' : '',
        noAudio: listItem.noAudio ? 'no audio' : '',
        isBest: !!listItem.isBest,
        noUpdate: 1
      };
      youtube.onPopupMenuItemClick(button, data);

      if (listItem.quality === 'ummy') {
        return;
      }

      if (listItem.forceDownload) {
        SaveFrom_Utils.downloadOnClick(e, null, listItem.useIframe || false );
      }
    },

    setButtonValue: (function() {
      var bindDlBtn = function(selectBtn, dlBtn, listItem) {
        var eIndex = dlBtn.dataset.eventInedx || 0;
        dlBtn.dataset.eventInedx = ++eIndex;
        dlBtn.addEventListener('click', function onClick(e){
          if (parseInt(this.dataset.eventInedx) !== eIndex) {
            this.removeEventListener('click', onClick);
            return;
          }
          youtube.onDlBtnClick.call(this, e, listItem, selectBtn);
        });

        dlBtn.removeAttribute('download');
        dlBtn.removeAttribute('title');

        if (!listItem) {
          return;
        }

        var title;
        if (listItem.quality === 'ummy') {
          mono.sendMessage({action: 'getUmmyIcon'}, function(dataImg) {
            selectBtn.insertBefore(mono.create('span', {
              style: {
                width: '16px',
                height: '16px',
                backgroundImage: 'url('+dataImg+')',
                backgroundRepeat: 'no-repeat',
                backgroundPosition: 'center',
                marginLeft: '6px',
                display: 'inline-block',
                verticalAlign: 'top'
              }
            }), selectBtn.lastChild);
          }, 'bg');
          title = mono.capitalize(listItem.quality);
        } else {
          var btnText = listItem.quality;
          if (!listItem.noVideo) {
            btnText = parseInt(btnText);
          }
          if (listItem['3d']) {
            btnText = '3D ' + btnText;
          }
          title = [listItem.format, btnText];
          if (listItem.noAudio) {
            title.push(language.withoutAudio);
          }
          title = title.join(' ');
          selectBtn.insertBefore(mono.create('span', {
            text: btnText,
            style: {
              marginLeft: '6px',
              verticalAlign: 'bottom'
            }
          }), selectBtn.lastChild);
        }

        dlBtn.title = title;
        dlBtn.href = listItem.href;

        if (listItem.title && listItem.format) {
          var ext = listItem.ext;
          if(!ext) {
            ext = listItem.format.toLowerCase();
          }
          dlBtn.setAttribute('download', SaveFrom_Utils.fileName.modify(listItem.title + '.' + ext) );
        }
      };

      var getBestItem = function(menuLinks) {
        var sList = [];
        for (var i = 0, item; item = menuLinks[i]; i++) {
          if (item.prop.noAudio || item.prop.noVideo) {
            continue;
          }
          if (item.prop.format === 'ummy') {
            continue;
          }
          if (!item.prop.isHidden) {
            sList.push(item.prop);
          }
        }
        menuLinks = sList;
        if (menuLinks.length > 0) {
          return menuLinks[0];
        }
        return undefined;
      };

      var prepMenuLinks = function(links, title) {
        var menuLinks = SaveFrom_Utils.popupMenu.prepareLinks.youtube(links, title);
        var ummyLink = undefined;

        var linkList = [];
        for (var i = 0, item; item = menuLinks[i]; i++) {
          if (item.quality === 'ummy') {
            ummyLink = item;
          }
          linkList.push({prop: item});
        }
        menuLinks = SaveFrom_Utils.popupMenu.sortMenuItems(linkList);
        var hasBest = false;
        menuLinks.forEach(function(item) {
          if (item.prop.isHidden) {
            return 1;
          }
          if (hasBest === false) {
            item.prop.isBest = hasBest = true;
            return 0;
          }
          return 0;
        });

        return {mL: menuLinks, uM: ummyLink};
      };

      var onGetLinks = function(selectBtn, dlBtn, links) {
        var i, item;
        if (!links || links.length === 0) {
          return bindDlBtn(selectBtn, dlBtn);
        }

        mono.storage.get('ytLastITag', function(storage) {
          var iTagLast = storage.ytLastITag;
          var title = youtube.getTitle();
          if(title) {
            title = modifyTitle(title);
          }

          var menuLinks = prepMenuLinks(links, title);
          var ummyLink = menuLinks.uM;
          menuLinks = menuLinks.mL;

          if (selectBtn.firstChild.tagName !== 'I') {
            selectBtn.removeChild(selectBtn.firstChild);
          }

          if (iTagLast === 'ummy' && ummyLink !== undefined) {
            bindDlBtn(selectBtn, dlBtn, ummyLink);
          } else {
            bindDlBtn(selectBtn, dlBtn, getBestItem(menuLinks));
          }
        });
      };

      return function(selectBtn) {
        var dlBtn = selectBtn.parentNode.firstChild;
        var vid = youtube.video_id;

        if(youtube.links[vid]) {
          return onGetLinks(selectBtn, dlBtn, youtube.links[vid]);
        }

        if(!youtube.swfargs || !youtube.token || !vid) {
          return onGetLinks(selectBtn, dlBtn, []);
        }

        bindDlBtn(selectBtn, dlBtn);

        var request = {
          action: 'getYoutubeLinks',
          extVideoId: vid,
          url: location.href
        };

        mono.sendMessage(request, function(response){
          if (response.links) {
            youtube.links[vid] = response.links;
          }
          return onGetLinks(selectBtn, dlBtn, response.links || []);
        }, 'bg');
      }
    })(),

    getLinks: function(event)
    {
      event.preventDefault();
      event.stopPropagation();
      var vid = youtube.video_id;

      var button = this;

      var d = button;
      while(d && d != document.body)
      {
        if(d.style.display == 'none')
          d.style.display = '';

        if(d.style.visibility == 'hidden')
          d.style.visibility = 'visible';

        d.style.zOrder = '9999';

        d = d.parentNode;
      }

      if(youtube.links[vid]) {
        youtube.showLinks(youtube.links[vid], button);
        return;
      }

      if(!youtube.swfargs || !youtube.token || !vid)
        return;


      var request = {
        action: 'getYoutubeLinks',
        extVideoId: vid,
        url: location.href
      };

      mono.sendMessage(request, function(response) {
        if (response.links) {
          youtube.links[vid] = response.links;
        }
        youtube.showLinks(response.links || [], button);
      }, 'bg');

      youtube.showLinks(language.download +
        ' ...', button);
    }
  };


  function modifyTitle(t)
  {
    t = t.replace(/[\x2F\x5C\x3A\x7C]/g, '-');
    t = t.replace(/[\x2A\x3F]/g, '');
    t = t.replace(/\x22/g, '\'');
    t = t.replace(/\x3C/g, '(');
    t = t.replace(/\x3E/g, ')');
    t = t.replace(/(?:^\s+)|(?:\s+$)/g, '');
    return t;
  }

  init();
});